# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import copy
import numpy as np
from math import ceil
from collections import OrderedDict
import logging


# +
class Layer:

    def __init__(self, name, size, layer_index, res=False):
        self.name = name
        self.size = []
        for x in size:
            self.size.append(x)
        self.layer_index = layer_index
        self.is_shortcut = True if 'downsample' in name else False
        if res:
            (
                self.stage,
                self.seq_index,
                self.block_index,
            ) = self._get_various_index_by_name(name)

    def _get_various_index_by_name(self, name):
        """Get the indices including stage, seq_ix, blk_ix.

        Same stage means the same feature map size.
        """
        global lastest_stage  # an awkward impel, just for now
        if name.startswith('module.'):
            name = name[
                7:]  # remove the prefix caused by pytorch data parallel

        if 'conv1' == name:  # TODO: this might not be so safe
            lastest_stage = 0
            return 0, None, None
        if ('linear' in name or 'fc' in name
            ):  # Note: this can be risky. Check it fully. TODO: @mingsun-tse
            return (
                lastest_stage + 1,
                None,
                None,
            )  # fc layer should always be the last layer
        else:
            try:
                stage = int(
                    name.split('.')[0][-1]
                )  # ONLY work for standard resnets. name example: layer2.2.conv1, layer4.0.downsample.0
                seq_ix = int(name.split('.')[1])
                if 'conv' in name.split('.')[-1]:
                    blk_ix = int(name[-1]) - 1
                else:
                    blk_ix = -1  # shortcut layer
                lastest_stage = stage
                return stage, seq_ix, blk_ix
            except:
                print('!Parsing the layer name failed: %s. Please check.' %
                      name)


class MetaPruner:

    def __init__(self, model, args, logger, passer):
        self.model = model
        self.args = args
        self.logger = logger
        self.test = lambda net: passer.test(passer.test_loader, net, passer.
                                            criterion, passer.args)
        self.train_loader = passer.train_loader
        self.criterion = passer.criterion
        self.save = passer.save
        self.is_single_branch = passer.is_single_branch

        self.layers = OrderedDict()
        self._register_layers()

        arch = self.args['arch']
        if arch.startswith('resnet'):
            # TODO: add block
            self.n_conv_within_block = 0
            if args['dataset'] == 'imagenet':
                if arch in ['resnet18', 'resnet34']:
                    self.n_conv_within_block = 2
                elif arch in ['resnet50', 'resnet101', 'resnet152']:
                    self.n_conv_within_block = 3
            else:
                self.n_conv_within_block = 2

        self.kept_wg = {}
        self.pruned_wg = {}
        self.get_pr()  # set up pr for each layer

    def _pick_pruned(self, w_abs, pr, mode='min'):
        if pr == 0:
            return []
        w_abs_list = w_abs.flatten()
        n_wg = len(w_abs_list)
        n_pruned = min(ceil(pr * n_wg), n_wg - 1)  # do not prune all
        if mode == 'rand':
            out = np.random.permutation(n_wg)[:n_pruned]
        elif mode == 'min':
            out = w_abs_list.sort()[1][:n_pruned]
        elif mode == 'max':
            out = w_abs_list.sort()[1][-n_pruned:]
        return out

    def _register_layers(self):
        """This will maintain a data structure that can return some useful
        information by the name of a layer."""
        ix = -1  # layer index, starts from 0
        max_len_name = 0
        layer_shape = {}
        for name, m in self.model.named_modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                if 'downsample' not in name:
                    ix += 1
                layer_shape[name] = [ix, m.weight.size()]
                max_len_name = max(max_len_name, len(name))

                size = m.weight.size()
                res = True if self.args['arch'].startswith('resnet') else False
                self.layers[name] = Layer(name, size, ix, res)

        max_len_ix = len('%s' % ix)
        print('Register layer index and kernel shape:')
        self.logger.info('Register layer index and kernel shape:')
        format_str = '[%{}d] %{}s -- kernel_shape: %s'.format(
            max_len_ix, max_len_name)
        for name, (ix, ks) in layer_shape.items():
            print(format_str % (ix, name, ks))
            self.logger.info(format_str % (ix, name, ks))

    def _next_conv(self, model, name, mm):
        if hasattr(self.layers[name], 'block_index'):
            block_index = self.layers[name].block_index
            if block_index == self.n_conv_within_block - 1:
                return None
        ix_conv = 0
        ix_mm = -1
        for n, m in model.named_modules():
            if isinstance(m, nn.Conv2d):
                ix_conv += 1
                if m == mm:
                    ix_mm = ix_conv
                if ix_mm != -1 and ix_conv == ix_mm + 1:
                    return n
        return None

    def _prev_conv(self, model, name, mm):
        if hasattr(self.layers[name], 'block_index'):
            block_index = self.layers[name].block_index
            if block_index in [
                    None,
                    0,
                    -1,
            ]:  # 1st conv, 1st conv in a block, 1x1 shortcut layer
                return None
        for n, _ in model.named_modules():
            if n in self.layers:
                ix = self.layers[n].layer_index
                if ix + 1 == self.layers[name].layer_index:
                    return n
        return None

    def _next_bn(self, model, mm):
        just_passed_mm = False
        for m in model.modules():
            if m == mm:
                just_passed_mm = True
            if just_passed_mm and isinstance(m, nn.BatchNorm2d):
                return m
        return None

    def _replace_module(self, model, name, new_m):
        """Replace the module <name> in <model> with <new_m> E.g.,
        'module.layer1.0.conv1' ==> model.__getattr__('module').__getattr__("la
        yer1").__getitem__(0).__setattr__('conv1', new_m)"""
        obj = model
        segs = name.split('.')
        for ix in range(len(segs)):
            s = segs[ix]
            if ix == len(segs) - 1:  # the last one
                if s.isdigit():
                    obj.__setitem__(int(s), new_m)
                else:
                    obj.__setattr__(s, new_m)
                return
            if s.isdigit():
                obj = obj.__getitem__(int(s))
            else:
                obj = obj.__getattr__(s)

    def _get_n_filter(self, model):
        """Do not consider the downsample 1x1 shortcuts."""
        n_filter = []
        for name, m in model.named_modules():
            if isinstance(m, nn.Conv2d):
                if not self.layers[name].is_shortcut:
                    n_filter.append(m.weight.size(0))
        return n_filter

    def _get_layer_pr_vgg(self, name):
        """Example: '[0-4:0.5, 5:0.6, 8-10:0.2]'
        6, 7 not mentioned, default value is 0
        """
        layer_index = self.layers[name].layer_index
        pr = self.args['stage_pr'][layer_index]
        if str(layer_index) in self.args['skip_layers']:
            pr = 0
        return pr

    def _get_layer_pr_resnet(self, name):
        """This function will determine the prune_ratio (pr) for each specific
        layer by a set of rules."""
        wg = self.args['wg']
        layer_index = self.layers[name].layer_index
        stage = self.layers[name].stage
        seq_index = self.layers[name].seq_index
        block_index = self.layers[name].block_index
        is_shortcut = self.layers[name].is_shortcut
        pr = self.args['stage_pr'][stage]

        # for unstructured pruning, no restrictions, every layer can be pruned
        if self.args['wg'] != 'weight':
            # do not prune the shortcut layers for now
            if is_shortcut:
                pr = 0

            # do not prune layers we set to be skipped
            layer_id = '%s.%s.%s' % (str(stage), str(seq_index),
                                     str(block_index))
            for s in self.args['skip_layers']:
                if s and layer_id.startswith(s):
                    pr = 0

            # for channel/filter prune, do not prune the 1st/last conv in a block
            if (wg == 'channel'
                    and block_index == 0) or (wg == 'filter' and block_index
                                              == self.n_conv_within_block - 1):
                pr = 0

        # Deprecated, will be removed:
        # # adjust accordingly if we explicitly provide the pr_ratio_file
        # if self.args.pr_ratio_file:
        #     line = open(self.args.pr_ratio_file).readline()
        #     pr_weight = strdict_to_dict(line, float)
        #     if str(layer_index) in pr_weight:
        #         pr = pr_weight[str(layer_index)] * pr
        return pr

    def get_pr(self):
        if self.is_single_branch(self.args['arch']):
            get_layer_pr = self._get_layer_pr_vgg
        else:
            get_layer_pr = self._get_layer_pr_resnet

        self.pr = {}
        if self.args[
                'stage_pr']:  # stage_pr may be None (in the case that base_pr_model is provided)
            for name, m in self.model.named_modules():
                if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                    self.pr[name] = get_layer_pr(name)
        else:
            assert self.args['base_pr_model']
            state = torch.load(self.args['base_pr_model'])
            self.pruned_wg_pr_model = state['pruned_wg']
            self.kept_wg_pr_model = state['kept_wg']
            for k in self.pruned_wg_pr_model:
                n_pruned = len(self.pruned_wg_pr_model[k])
                n_kept = len(self.kept_wg_pr_model[k])
                self.pr[k] = float(n_pruned) / (n_pruned + n_kept)
            print(
                "==> Load base_pr_model successfully and inherit its pruning ratio: '{}'"
                .format(self.args['base_pr_model']))

            self.logger.info(
                "==> Load base_pr_model successfully and inherit its pruning ratio: '{}'"
                .format(self.args['base_pr_model']))

    def _get_kept_wg_L1(self):
        if self.args['base_pr_model'] and self.args[
                'inherit_pruned'] == 'index':
            self.pruned_wg = self.pruned_wg_pr_model
            self.kept_wg = self.kept_wg_pr_model
            print(
                "==> Inherit the pruned index from base_pr_model: '{}'".format(
                    self.args['base_pr_model']))

            self.logger.info(
                "==> Inherit the pruned index from base_pr_model: '{}'".format(
                    self.args['base_pr_model']))
        else:
            wg = self.args['wg']
            for name, m in self.model.named_modules():
                if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                    shape = m.weight.data.shape
                    if wg == 'filter':
                        score = (m.weight.abs().mean(dim=[1, 2, 3])
                                 if len(shape) == 4 else m.weight.abs().mean(
                                     dim=1))
                    elif wg == 'channel':
                        score = (m.weight.abs().mean(dim=[0, 2, 3])
                                 if len(shape) == 4 else m.weight.abs().mean(
                                     dim=0))
                    elif wg == 'weight':
                        score = m.weight.abs().flatten()
                    else:
                        raise NotImplementedError
                    self.pruned_wg[name] = self._pick_pruned(
                        score, self.pr[name], self.args['pick_pruned'])
                    self.kept_wg[name] = [
                        i for i in range(len(score))
                        if i not in self.pruned_wg[name]
                    ]
                    logtmp = '[%2d %s] got pruned wg by L1 sorting (%s), pr %.4f' % (
                        self.layers[name].layer_index,
                        name,
                        self.args['pick_pruned'],
                        self.pr[name],
                    )

                    # compare the pruned weights picked by L1-sorting vs. other criterion which provides the base_pr_model (e.g., OBD)
                    if self.args['base_pr_model']:
                        intersection = [
                            x for x in self.pruned_wg_pr_model[name]
                            if x in self.pruned_wg[name]
                        ]
                        intersection_ratio = (len(intersection) /
                                              len(self.pruned_wg[name]) if len(
                                                  self.pruned_wg[name]) else 0)
                        logtmp += (
                            ', intersection ratio of the weights picked by L1 vs. base_pr_model: %.4f (%d)'
                            % (intersection_ratio, len(intersection)))
                    self.logger.info(logtmp)

    def _get_kept_filter_channel(self, m, name):
        if self.args['wg'] == 'channel':
            kept_chl = self.kept_wg[name]
            next_conv = self._next_conv(self.model, name, m)
            if not next_conv:
                kept_filter = range(m.weight.size(0))
            else:
                kept_filter = self.kept_wg[next_conv]

        elif self.args['wg'] == 'filter':
            kept_filter = self.kept_wg[name]
            prev_conv = self._prev_conv(self.model, name, m)
            if not prev_conv:
                kept_chl = range(m.weight.size(1))
            else:
                kept_chl = self.kept_wg[prev_conv]

        return kept_filter, kept_chl

    def _prune_and_build_new_model(self):
        if self.args['wg'] == 'weight':
            self._get_masks()
            return

        new_model = copy.deepcopy(self.model)
        cnt_linear = 0
        for name, m in self.model.named_modules():
            if isinstance(m, nn.Conv2d):
                kept_filter, kept_chl = self._get_kept_filter_channel(m, name)

                # copy conv weight and bias
                bias = False if isinstance(m.bias, type(None)) else True
                kept_weights = m.weight.data[kept_filter][:, kept_chl, :, :]
                new_conv = nn.Conv2d(
                    kept_weights.size(1),
                    kept_weights.size(0),
                    m.kernel_size,
                    m.stride,
                    m.padding,
                    m.dilation,
                    m.groups,
                    bias,
                ).cuda()
                new_conv.weight.data.copy_(
                    kept_weights)  # load weights into the new module
                if bias:
                    kept_bias = m.bias.data[kept_filter]
                    new_conv.bias.data.copy_(kept_bias)

                # load the new conv
                self._replace_module(new_model, name, new_conv)

                # get the corresponding bn (if any) for later use
                next_bn = self._next_bn(self.model, m)

            elif isinstance(m, nn.BatchNorm2d) and m == next_bn:
                new_bn = nn.BatchNorm2d(
                    len(kept_filter),
                    eps=m.eps,
                    momentum=m.momentum,
                    affine=m.affine,
                    track_running_stats=m.track_running_stats,
                ).cuda()

                # copy bn weight and bias
                if self.args['copy_bn_w']:
                    weight = m.weight.data[kept_filter]
                    new_bn.weight.data.copy_(weight)
                if self.args['copy_bn_b']:
                    bias = m.bias.data[kept_filter]
                    new_bn.bias.data.copy_(bias)

                # copy bn running stats
                new_bn.running_mean.data.copy_(m.running_mean[kept_filter])
                new_bn.running_var.data.copy_(m.running_var[kept_filter])
                new_bn.num_batches_tracked.data.copy_(m.num_batches_tracked)

                # load the new bn
                self._replace_module(new_model, name, new_bn)

            elif isinstance(m, nn.Linear):
                cnt_linear += 1
                if cnt_linear == 1:
                    # get the last conv
                    last_conv = ''
                    last_conv_name = ''
                    for n, mm in self.model.named_modules():
                        if isinstance(mm, nn.Conv2d):
                            last_conv = mm
                            last_conv_name = n
                    kept_filter_last_conv, _ = self._get_kept_filter_channel(
                        last_conv, last_conv_name)

                    # get kept weights
                    dim_in = m.weight.size(1)
                    fm_size = int(dim_in /
                                  last_conv.weight.size(0))  # 36 for alexnet
                    kept_dim_in = []
                    for i in kept_filter_last_conv:
                        tmp = list(range(i * fm_size, i * fm_size + fm_size))
                        kept_dim_in += tmp
                    kept_weights = m.weight.data[:, kept_dim_in]

                    # build the new linear layer
                    bias = False if isinstance(m.bias, type(None)) else True
                    new_linear = nn.Linear(
                        in_features=len(kept_dim_in),
                        out_features=m.out_features,
                        bias=bias,
                    ).cuda()
                    new_linear.weight.data.copy_(kept_weights)
                    if bias:
                        new_linear.bias.data.copy_(m.bias.data)

                    # load the new linear
                    self._replace_module(new_model, name, new_linear)

        self.model = new_model
        print('MODEL PRUNING COMPLETE : META PRUNER')
        self.logger.info('MODEL PRUNING COMPLETE : META PRUNER')
        n_filter = self._get_n_filter(self.model)
        print(n_filter)

    def _get_masks(self):
        """Get masks for unstructured pruning."""
        self.mask = {}
        for name, m in self.model.named_modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                mask = torch.ones_like(m.weight.data).cuda().flatten()
                pruned = self.pruned_wg[name]
                mask[pruned] = 0
                self.mask[name] = mask.view_as(m.weight.data)
        print('Get masks done for weight pruning')
        self.logger.info('Get masks done for weight pruning')



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.optim as optim
import os, copy, time, pickle, numpy as np, math
import torchvision.models as models
from .meta_pruner import MetaPruner
import matplotlib.pyplot as plt
import logging

from ..pns import SlimPruner
from ..utils import summary_model
from trailmet.utils import save_checkpoint


class LitModel(nn.Module):

    def __init__(self, args):
        super(LitModel, self).__init__()
        model = getattr(models, args['arch'])(num_classes=args['num_classes'])
        # to get better result on cifar10
        model.conv1 = torch.nn.Conv2d(3,
                                      64,
                                      kernel_size=3,
                                      stride=1,
                                      padding=1,
                                      bias=False)
        model.maxpool = torch.nn.Identity()
        self.model = model
        # self.model.load_state_dict(torch.load('baseline_r50_t200.pth')['state_dict'])

    def forward(self, x):
        return self.model(x)


# +
class Pruner(MetaPruner):

    def __init__(self, model, args, logger, passer):
        super(Pruner, self).__init__(model, args, logger, passer)

        # Reg related variables
        self.reg = {}
        self.delta_reg = {}
        self.hist_mag_ratio = {}
        self.n_update_reg = {}
        self.iter_update_reg_finished = {}
        self.iter_finish_pick = {}
        self.iter_stabilize_reg = math.inf
        self.original_w_mag = {}
        self.original_kept_w_mag = {}
        self.ranking = {}
        self.pruned_wg_L1 = {}
        self.all_layer_finish_pick = False
        self.w_abs = {}
        self.smv = {}
        self.mag_reg_log = {}
        self.logger = logger
        if self.args[
                'method'] == 'GReg-2':  # AdaReg is the old name for GReg-2
            self.original_model = copy.deepcopy(self.model)

        # prune_init, to determine the pruned weights
        # this will update the 'self.kept_wg' and 'self.pruned_wg'
        if self.args['method'] in ['GReg-1', 'GReg-2']:
            self._get_kept_wg_L1()
        for k, v in self.pruned_wg.items():
            self.pruned_wg_L1[k] = v
        if (
                self.args['method'] == 'GReg-2'
        ):  # GReg-2 will determine which wgs to prune later, so clear it here
            self.kept_wg = {}
            self.pruned_wg = {}

        self.prune_state = 'update_reg'
        for name, m in self.model.named_modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                shape = m.weight.data.shape

                # initialize reg
                if self.args['wg'] == 'weight':
                    self.reg[name] = torch.zeros_like(
                        m.weight.data).flatten().cuda()
                else:
                    self.reg[name] = torch.zeros(shape[0], shape[1]).cuda()

                # get original weight magnitude
                w_abs = self._get_score(m)
                n_wg = len(w_abs)
                self.ranking[name] = []
                for _ in range(n_wg):
                    self.ranking[name].append([])
                self.original_w_mag[name] = m.weight.abs().mean().item()
                kept_wg_L1 = [
                    i for i in range(n_wg) if i not in self.pruned_wg_L1[name]
                ]
                self.original_kept_w_mag[name] = w_abs[kept_wg_L1].mean().item(
                )

    def _pick_pruned_wg(self, w, pr):
        if pr == 0:
            return []
        elif pr > 0:
            w = w.flatten()
            n_pruned = min(math.ceil(pr * w.size(0)),
                           w.size(0) - 1)  # do not prune all
            return w.sort()[1][:n_pruned]
        elif pr == -1:  # automatically decide lr by each layer itself
            tmp = w.flatten().sort()[0]
            n_not_consider = int(len(tmp) * 0.02)
            w = tmp[n_not_consider:-n_not_consider]

            sorted_w, sorted_index = w.flatten().sort()
            max_gap = 0
            max_index = 0
            for i in range(len(sorted_w) - 1):
                # gap = sorted_w[i+1:].mean() - sorted_w[:i+1].mean()
                gap = sorted_w[i + 1] - sorted_w[i]
                if gap > max_gap:
                    max_gap = gap
                    max_index = i
            max_index += n_not_consider
            return sorted_index[:max_index + 1]
        else:
            self.logger.info('Wrong pr. Please check.')
            exit(1)

    def _update_mag_ratio(self, m, name, w_abs, pruned=None):
        if type(pruned) == type(None):
            pruned = self.pruned_wg[name]
        kept = [i for i in range(len(w_abs)) if i not in pruned]
        ave_mag_pruned = w_abs[pruned].mean()
        ave_mag_kept = w_abs[kept].mean()
        if len(pruned):
            mag_ratio = ave_mag_kept / ave_mag_pruned
            if name in self.hist_mag_ratio:
                self.hist_mag_ratio[name] = (self.hist_mag_ratio[name] * 0.9 +
                                             mag_ratio * 0.1)
            else:
                self.hist_mag_ratio[name] = mag_ratio
        else:
            mag_ratio = math.inf
            self.hist_mag_ratio[name] = math.inf

        # print
        mag_ratio_now_before = ave_mag_kept / self.original_kept_w_mag[name]
        if self.total_iter % self.args['print_interval'] == 0:
            self.logger.info('    mag_ratio %.4f mag_ratio_momentum %.4f' %
                             (mag_ratio, self.hist_mag_ratio[name]))
            self.logger.info(
                '    for kept weights, original_kept_w_mag %.6f, now_kept_w_mag %.6f ratio_now_over_original %.4f'
                % (self.original_kept_w_mag[name], ave_mag_kept,
                   mag_ratio_now_before))
        return mag_ratio_now_before

    def _get_score(self, m):
        shape = m.weight.data.shape
        if self.args['wg'] == 'channel':
            w_abs = (m.weight.abs().mean(dim=[0, 2, 3])
                     if len(shape) == 4 else m.weight.abs().mean(dim=0))
        elif self.args['wg'] == 'filter':
            w_abs = (m.weight.abs().mean(dim=[1, 2, 3])
                     if len(shape) == 4 else m.weight.abs().mean(dim=1))
        elif self.args['wg'] == 'weight':
            w_abs = m.weight.abs().flatten()
        return w_abs

    def _fix_reg(self, m, name):
        if self.pr[name] == 0:
            return True
        if self.args['wg'] != 'weight':
            self._update_mag_ratio(m, name, self.w_abs[name])

        pruned = self.pruned_wg[name]
        if self.args['wg'] == 'channel':
            self.reg[name][:, pruned] = self.args['reg_upper_limit']
        elif self.args['wg'] == 'filter':
            self.reg[name][pruned, :] = self.args['reg_upper_limit']
        elif self.args['wg'] == 'weight':
            self.reg[name][pruned] = self.args['reg_upper_limit']

        finish_update_reg = self.total_iter > self.args['fix_reg_interval']
        return finish_update_reg

    def _greg_1(self, m, name):
        if self.pr[name] == 0:
            return True

        if self.args['wg'] != 'weight':  # weight is too slow
            self._update_mag_ratio(m, name, self.w_abs[name])

        pruned = self.pruned_wg[name]
        if self.args['wg'] == 'channel':
            self.reg[name][:, pruned] += self.args['reg_granularity_prune']
        elif self.args['wg'] == 'filter':
            self.reg[name][pruned, :] += self.args['reg_granularity_prune']
        elif self.args['wg'] == 'weight':
            self.reg[name][pruned] += self.args['reg_granularity_prune']
        else:
            raise NotImplementedError

        # when all layers are pushed hard enough, stop
        if (
                self.args['wg'] == 'weight'
        ):  # for weight, do not use the magnitude ratio condition, because 'hist_mag_ratio' is not updated, too costly
            finish_update_reg = False
        else:
            finish_update_reg = True
            for k in self.hist_mag_ratio:
                if self.hist_mag_ratio[k] < self.args['mag_ratio_limit']:
                    finish_update_reg = False
        self.smv[name] = self.reg[name].max()
        return finish_update_reg or self.reg[name].max(
        ) > self.args['reg_upper_limit']

    def _greg_2(self, m, name):
        layer_index = self.layers[name].layer_index
        w_abs = self.w_abs[name]
        n_wg = len(w_abs)
        pr = self.pr[name]
        if pr == 0:
            self.kept_wg[name] = range(n_wg)
            self.pruned_wg[name] = []
            self.iter_finish_pick[name] = self.total_iter
            return True

        if name in self.iter_finish_pick:
            recover_reg = self.args['reg_granularity_recover']
            # for pruned weights, push them more
            if self.args['wg'] == 'channel':
                self.reg[name][:, self.pruned_wg[name]] += self.args[
                    'reg_granularity_prune']
                self.reg[name][:, self.kept_wg[name]] = recover_reg
            elif self.args['wg'] == 'filter':
                self.reg[name][self.pruned_wg[name], :] += self.args[
                    'reg_granularity_prune']
                self.reg[name][self.kept_wg[name], :] = recover_reg
            elif self.args['wg'] == 'weight':
                self.reg[name][
                    self.pruned_wg[name]] += self.args['reg_granularity_prune']
                self.reg[name][self.kept_wg[name]] = recover_reg

            # for kept weights, bring them back
            # 09/22 update: It seems negative reg is a bad idea to bring back magnitude.
            """
            current_w_mag = w_abs[self.kept_wg[name]].mean()
            recover_reg = (current_w_mag / self.original_kept_w_mag[name] - 1).item() \
                * self.args["weight_decay"] * self.args["reg_multiplier"] * 10
            if recover_reg > 0:
                recover_reg = 0
            if self.args["wg"] == 'channel':
                self.reg[name][:, self.kept_wg[name]] = recover_reg
            elif self.args["wg"] == 'filter':
                self.reg[name][self.kept_wg[name], :] = recover_reg
            """
            if self.total_iter % self.args['print_interval'] == 0:
                self.logger.info(
                    '    prune stage, push the pruned (reg = %.5f) to zero; for kept weights, reg = %.5f'
                    % (self.reg[name].max().item(), recover_reg))

        else:
            self.reg[name] += self.args['reg_granularity_pick']

        # plot w_abs distribution
        if self.total_iter % self.args['plot_interval'] == 0:
            self._plot_mag_ratio(w_abs, name)
        if self.total_iter % self.args['plot_interval'] == 0:
            self._log_down_mag_reg(w_abs, name)

        # save order
        # if not hasattr(self, 'order_by_L1'):
        # self.order_by_L1 = {}
        # self.wg_preprune = {} # deprecated: dict costs too much memory and seriously slow down training
        # if name in self.order_by_L1:
        #     self.order_by_L1[name] = torch.cat([self.order_by_L1[name], order], dim=0)
        #     self.wg_preprune[name] = torch.cat([self.wg_preprune[name], wg_pruned], dim=0)
        # else:
        #     self.order_by_L1[name] = order
        #     self.wg_preprune[name] = wg_pruned
        if (self.args['save_order_log']
                and self.total_iter % self.args['update_reg_interval'] == 0):
            if not hasattr(self, 'order_log'):
                self.order_log = open(
                    '%s/order_log.txt' % self.logger.log_path, 'w+')

            order = w_abs.argsort().argsort()
            n_pruned = min(math.ceil(pr * n_wg), n_wg - 1)
            wg_pruned = w_abs.argsort()[:n_pruned]

            order = [str(x.item()) for x in order]  # for each wg, its ranking
            wg_pruned = [str(x.item()) for x in wg_pruned]

            logtmp = ' '.join(order)
            logtmp = 'Iter %d Layer#%d %s order_by_L1 %s' % (
                self.total_iter,
                layer_index,
                name,
                logtmp,
            )
            print(logtmp, file=self.order_log)

            logtmp = ' '.join(wg_pruned)
            logtmp = 'Iter %d Layer#%d %s wg_preprune %s' % (
                self.total_iter,
                layer_index,
                name,
                logtmp,
            )
            print(logtmp, file=self.order_log, flush=True)

        # print to check magnitude ratio
        if self.args['wg'] != 'weight':
            if name in self.iter_finish_pick:
                self._update_mag_ratio(m, name, w_abs)
            else:
                pruned_wg = self._pick_pruned_wg(w_abs, pr)
                self._update_mag_ratio(m, name, w_abs,
                                       pruned=pruned_wg)  # just print to check

        # check if picking finishes
        finish_pick_cond = self.reg[name].max(
        ) >= self.args['reg_upper_limit_pick']
        if name not in self.iter_finish_pick and finish_pick_cond:
            self.iter_finish_pick[name] = self.total_iter
            pruned_wg = self._pick_pruned_wg(w_abs, pr)
            kept_wg = [i for i in range(n_wg) if i not in pruned_wg]
            self.kept_wg[name] = kept_wg
            self.pruned_wg[name] = pruned_wg
            picked_wg_in_common = [
                i for i in pruned_wg if i in self.pruned_wg_L1[name]
            ]
            common_ratio = (len(picked_wg_in_common) /
                            len(pruned_wg) if len(pruned_wg) else -1)
            n_finish_pick = len(self.iter_finish_pick)
            self.logger.info(
                '    [%d] just finished pick (n_finish_pick = %d). %.2f in common chosen by L1 & GReg-2. Iter = %d'
                % (layer_index, n_finish_pick, common_ratio, self.total_iter))

            # re-scale the weights to recover the response magnitude
            # factor = self.original_w_mag[name] / m.weight.abs().mean()
            # m.weight.data.mul_(factor)
            # self.logger.info('    rescale weight by %.4f' % factor.item())

            # check if all layers finish picking
            self.all_layer_finish_pick = True
            for k in self.reg:
                if self.pr[k] > 0 and (k not in self.iter_finish_pick):
                    self.all_layer_finish_pick = False
                    break

        # save mag_reg_log
        if self.args['save_mag_reg_log'] and (
                self.total_iter % self.args['save_interval'] == 0
                or name in self.iter_finish_pick):
            out = os.path.join(self.logger.log_path,
                               '%d_mag_reg_log.npy' % layer_index)
            np.save(out, self.mag_reg_log[name])
            if self.all_layer_finish_pick:
                exit(0)

        if self.args['method'] == 'GReg-2':
            finish_update_reg = False
        else:
            cond0 = name in self.iter_finish_pick  # finished picking
            if self.args['wg'] == 'weight':
                cond1 = self.reg[name].max() > self.args['reg_upper_limit']
            else:
                cond1 = (
                    self.hist_mag_ratio[name] >= self.args['mag_ratio_limit']
                    or self.reg[name].max() > self.args['reg_upper_limit'])
            finish_update_reg = cond0 and cond1
        return finish_update_reg

    def _update_reg(self):
        for name, m in self.model.named_modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                cnt_m = self.layers[name].layer_index
                pr = self.pr[name]

                if name in self.iter_update_reg_finished.keys():
                    continue

                if self.total_iter % self.args['print_interval'] == 0:
                    self.logger.info(
                        "[%d] Update reg for layer '%s'. Pr = %s. Iter = %d" %
                        (cnt_m, name, pr, self.total_iter))

                # get the importance score (L1-norm in this case)
                self.w_abs[name] = self._get_score(m)

                # update reg functions, two things:
                # (1) update reg of this layer (2) determine if it is time to stop update reg
                if self.args['method'] == 'FixReg':
                    finish_update_reg = self._fix_reg(m, name)
                elif self.args['method'] == 'GReg-1':
                    finish_update_reg = self._greg_1(m, name)
                elif self.args['method'] == 'GReg-2':
                    finish_update_reg = self._greg_2(m, name)
                else:
                    self.logger.info(
                        "Wrong '--method' argument, please check.")
                    exit(1)

                # check prune state
                if finish_update_reg:
                    # after 'update_reg' stage, keep the reg to stabilize weight magnitude
                    self.iter_update_reg_finished[name] = self.total_iter
                    self.logger.info(
                        "==> [%d] Just finished 'update_reg'. Iter = %d" %
                        (cnt_m, self.total_iter))

                    # check if all layers finish 'update_reg'
                    self.prune_state = 'stabilize_reg'
                    for n, mm in self.model.named_modules():
                        if isinstance(mm, nn.Conv2d) or isinstance(
                                mm, nn.Linear):
                            if n not in self.iter_update_reg_finished:
                                self.prune_state = 'update_reg'
                                break
                    if self.prune_state == 'stabilize_reg':
                        self.iter_stabilize_reg = self.total_iter
                        self.logger.info(
                            "==> All layers just finished 'update_reg', go to 'stabilize_reg'. Iter = %d"
                            % self.total_iter)
                        self._save_model(mark='just_finished_update_reg')

                # after reg is updated, print to check
                if self.total_iter % self.args['print_interval'] == 0:
                    self.logger.info(
                        '    reg_status: min = %.5f ave = %.5f max = %.5f' % (
                            self.reg[name].min(),
                            self.reg[name].mean(),
                            self.reg[name].max(),
                        ))

    def _apply_reg(self):
        for name, m in self.model.named_modules():
            if name in self.reg:
                reg = self.reg[name]  # [N, C]
                if self.args['wg'] in ['filter', 'channel']:
                    if reg.shape != m.weight.data.shape:
                        reg = reg.unsqueeze(2).unsqueeze(3)  # [N, C, 1, 1]
                elif self.args['wg'] == 'weight':
                    reg = reg.view_as(m.weight.data)  # [N, C, H, W]
                l2_grad = reg * m.weight
                if self.args['block_loss_grad']:
                    m.weight.grad = l2_grad
                else:
                    m.weight.grad += l2_grad

    def _resume_prune_status(self, ckpt_path):
        state = torch.load(ckpt_path)
        self.model = state['model'].cuda()
        self.model.load_state_dict(state['state_dict'])
        self.optimizer = optim.SGD(
            self.model.parameters(),
            lr=self.args['lr_pick']
            if self.args['__dict__'].get('AdaReg_only_picking') else
            self.args['lr_prune'],
            momentum=self.args['momentum'],
            weight_decay=self.args['weight_decay'],
        )
        self.optimizer.load_state_dict(state['optimizer'])
        self.prune_state = state['prune_state']
        self.total_iter = state['iter']
        self.iter_stabilize_reg = state.get('iter_stabilize_reg', math.inf)
        self.reg = state['reg']
        self.hist_mag_ratio = state['hist_mag_ratio']

    def _save_model(self, acc1=0, acc5=0, mark=''):
        state = {
            'iter': self.total_iter,
            'prune_state': self.prune_state,  # we will resume prune_state
            'arch': self.args['arch'],
            'model': self.model,
            'state_dict': self.model.state_dict(),
            'iter_stabilize_reg': self.iter_stabilize_reg,
            'acc1': acc1,
            'acc5': acc5,
            'optimizer': self.optimizer.state_dict(),
            'reg': self.reg,
            'hist_mag_ratio': self.hist_mag_ratio,
        }
        save_checkpoint(
            state,
            is_best=False,
            save='./checkpoints',
            file_name='just_finished_update_reg',
        )

    def prune(self):
        self.model = self.model.train()
        self.optimizer = optim.SGD(
            self.model.parameters(),
            lr=self.args['lr_pick']
            if self.args['method'] == 'GReg-2' else self.args['lr_prune'],
            momentum=self.args['momentum'],
            weight_decay=self.args['weight_decay'],
        )

        # resume model, optimzer, prune_status
        self.total_iter = -1
        if self.args['resume_path']:
            self._resume_prune_status(self.args['resume_path'])
            self._get_kept_wg_L1(
            )  # get pruned and kept wg from the resumed model
            self.model = self.model.train()
            self.logger.info(
                "Resume model successfully: '{}'. Iter = {}. prune_state = {}".
                format(self.args['resume_path'], self.total_iter,
                       self.prune_state))

        t1 = time.time()
        acc1 = acc5 = 0
        while True:
            for _, (inputs, targets) in enumerate(self.train_loader):
                inputs, targets = inputs.cuda(), targets.cuda()
                self.total_iter += 1
                total_iter = self.total_iter

                #                 # test
                #                 if total_iter % self.args["test_interval"] == 0:
                #                     acc1, acc5, *_ = self.test(self.model)
                #                     self.accprint("Acc1 = %.4f Acc5 = %.4f Iter = %d (before update) [prune_state = %s, method = %s]" %
                #                         (acc1, acc5, total_iter, self.prune_state, self.args["method"]))

                #                 # save model (save model before a batch starts)
                #                 if total_iter % self.args["save_interval"] == 0:
                #                     self._save_model(acc1, acc5)
                #                     self.logger.info('Periodically save model done. Iter = {}'.format(total_iter))

                #                 if total_iter % self.args["print_interval"] == 0:
                #                     self.logger.info("")
                #                     self.logger.info("Iter = %d [prune_state = %s, method = %s] "
                #                         % (total_iter, self.prune_state, self.args["method"]) + "-"*40)

                # forward
                self.model.train()
                y_ = self.model(inputs)
                print(self.prune_state, total_iter,
                      self.args['update_reg_interval'])
                if (self.prune_state == 'update_reg' and
                        total_iter % self.args['update_reg_interval'] == 0):
                    self._update_reg()

                # normal training forward
                loss = self.criterion(y_, targets)
                self.optimizer.zero_grad()
                loss.backward()

                # after backward but before update, apply reg to the grad
                self._apply_reg()
                self.optimizer.step()
                if self.total_iter % 2000 == 0:
                    print(self.total_iter)
                    print(self.smv)
                #                 if(self.total_iter == 30): I added this
                #                     self.prune_state = 'stabilize_reg'
                #         total_iter = self.iter_stabilize_reg+self.args["stabilize_reg_interval"]
                # log print
                if total_iter % self.args['print_interval'] == 0:
                    w_abs_sum = 0
                    w_num_sum = 0
                    cnt_m = 0
                    for _, m in self.model.named_modules():
                        if isinstance(m, nn.Conv2d):
                            cnt_m += 1
                            w_abs_sum += m.weight.abs().sum()
                            w_num_sum += m.weight.numel()

                    _, predicted = y_.max(1)
                    correct = predicted.eq(targets).sum().item()
                    train_acc = correct / targets.size(0)
                    self.logger.info(
                        'After optim update, ave_abs_weight: %.10f current_train_loss: %.4f current_train_acc: %.4f'
                        % (w_abs_sum / w_num_sum, loss.item(), train_acc))

                # Save heatmap of weights to check the magnitude
                # if total_iter % self.args["plot_interval"] == 0:
                #     cnt_m = 0
                #     for m in self.modules:
                #         cnt_m += 1
                #         if isinstance(m, nn.Conv2d):
                #             out_path1 = os.path.join(self.logger.logplt_path, "m%d_iter%d_weights_heatmap.jpg" %
                #                     (cnt_m, total_iter))
                #             out_path2 = os.path.join(self.logger.logplt_path, "m%d_iter%d_reg_heatmap.jpg" %
                #                     (cnt_m, total_iter))
                #             plot_weights_heatmap(m.weight.mean(dim=[2, 3]), out_path1)
                #             plot_weights_heatmap(self.reg[name], out_path2)

                if self.args[
                        'method'] == 'GReg-1' and self.all_layer_finish_pick:
                    self.logger.info(
                        'GReg-2 just finished picking for all layers. Resume original model and switch to GReg-1. Iter = %d'
                        % total_iter)

                    # save picked wg
                    pkl_path = os.path.join('./logs', 'picked_wg.pkl')
                    with open(pkl_path, 'wb') as f:
                        pickle.dump(self.pruned_wg, f)
                    exit(0)

                    # set to GReg-1 method
                    self.model = self.original_model  # reload the original model
                    self.optimizer = optim.SGD(
                        self.model.parameters(),
                        lr=self.args['lr_prune'],
                        momentum=self.args['momentum'],
                        weight_decay=self.args['weight_decay'],
                    )
                    self.args['method'] = 'GReg-1'
                    self.args[
                        'AdaReg_only_picking'] = False  # do not get in again
                    # reinit
                    for k in self.reg:
                        self.reg[k] = torch.zeros_like(self.reg[k]).cuda()
                    self.hist_mag_ratio = {}

                if self.args[
                        'method'] == 'GReg-1' and self.all_layer_finish_pick:
                    self._prune_and_build_new_model()
                    self.logger.info(
                        "GReg-2 just finished picking for all layers. Pruned and go to 'finetune'. Iter = %d"
                        % total_iter)
                    return copy.deepcopy(self.model)

                # change prune state
                if (self.prune_state == 'stabilize_reg'
                        and total_iter - self.iter_stabilize_reg
                        == self.args['stabilize_reg_interval']):
                    #                   self._prune_and_build_new_model()
                    #                   self.logger.info("'stabilize_reg' is done. Pruned, go to 'finetune'. Iter = %d" % total_iter)
                    new_model = LitModel(self.args)
                    A1 = copy.deepcopy(self.model.state_dict())
                    L1 = list(A1.keys())
                    for key in L1:
                        new_key = str('model.' + key)
                        A1[new_key] = A1.pop(key)
                    new_model.load_state_dict(A1)
                    self.model = new_model
                    if self.args['arch'] == 'resnet50':
                        schema = './schema/resnet50.json'
                    elif self.args['arch'] == 'mobilenet_v2':
                        schema = './schema/mobilenet_v2.json'
                    pruner = SlimPruner(self.model, schema)
                    pruning_result = pruner.run(self.args['prune_ratio'])
                    print('Summary of pruned_model')
                    summary_model(pruner.pruned_model)

                    return pruning_result, pruner.pruned_model
                #                 if self.prune_state == "stabilize_reg":
                #                     self._prune_and_build_new_model()
                #                     #print(self.model)
                #                     self.logger.info("'stabilize_reg' is done. Pruned, go to 'finetune'. Iter = %d" % total_iter)
                #                     return copy.deepcopy(self.model)

                if total_iter % self.args['print_interval'] == 0:
                    t2 = time.time()
                    total_time = t2 - t1
                    self.logger.info(
                        'speed = %.4f iter/s' %
                        (self.args['print_interval'] / total_time))
                    t1 = t2

    def _plot_mag_ratio(self, w_abs, name):
        fig, ax = plt.subplots()
        max_ = w_abs.max().item()
        w_abs_normalized = (w_abs / max_).data.cpu().numpy()
        ax.plot(w_abs_normalized)
        ax.set_ylim([0, 1])
        ax.set_xlabel('filter index')
        ax.set_ylabel('relative L1-norm ratio')
        layer_index = self.layers[name].layer_index
        shape = self.layers[name].size
        ax.set_title('layer %d iter %d shape %s\n(max = %s)' %
                     (layer_index, self.total_iter, shape, max_))
        out = os.path.join(
            self.logger.logplt_path,
            '%d_iter%d_w_abs_dist.jpg' % (layer_index, self.total_iter),
        )
        fig.savefig(out)
        plt.close(fig)
        np.save(out.replace('.jpg', '.npy'), w_abs_normalized)

    def _log_down_mag_reg(self, w_abs, name):
        step = self.total_iter
        reg = self.reg[name].max().item()
        mag = w_abs.data.cpu().numpy()
        if name not in self.mag_reg_log:
            values = [[step, reg, mag]]
            log = {
                'name': name,
                'layer_index': self.layers[name].layer_index,
                'shape': self.layers[name].size,
                'values': values,
            }
            self.mag_reg_log[name] = log
        else:
            values = self.mag_reg_log[name]['values']
            values.append([step, reg, mag])



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from .bitsplit import BitSplit
from .brecq import BRECQ
from .lapq import LAPQ
from .methods import (
    UniformAffineQuantizer,
    AdaRoundQuantizer,
    BitSplitQuantizer,
    ActQuantizer,
    QuantizationBase,
    UniformQuantization,
    ClippedUniformQuantization,
    FixedClipValueQuantization,
    MaxAbsStaticQuantization,
    LearnedStepSizeQuantization,
    LpNormQuantization,
)
from .qmodel import (
    QuantBasicBlock,
    QuantBottleneck,
    QuantInvertedResidual,
    QuantModule,
    BaseQuantBlock,
    QBasicBlock,
    QBottleneck,
    QInvertedResidual,
    ActivationModuleWrapper,
    ParameterModuleWrapper,
)
from .quantize import (
    BaseQuantization,
    StraightThrough,
    RoundSTE,
    Conv2dFunctor,
    LinearFunctor,
    FoldBN,
)
from .reconstruct import (
    StopForwardException,
    DataSaverHook,
    GetLayerInpOut,
    save_inp_oup_data,
    GradSaverHook,
    GetLayerGrad,
    save_grad_data,
    LinearTempDecay,
    LayerLossFunction,
    layer_reconstruction,
    BlockLossFunction,
    block_reconstruction,
)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.utils
import torch.backends.cudnn as cudnn
import math
import shutil
from pathlib import Path
from collections import OrderedDict
from thop import profile
import time, datetime
import logging
import torch.utils.data.distributed
from torch.cuda.amp import autocast, GradScaler
from trailmet.algorithms.prune.prune import BasePruning
from trailmet.models.resnet_chip import resnet_50

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import (
    AverageMeter,
    accuracy,
    save_checkpoint,
    CrossEntropyLabelSmooth,
    extract_sparsity,
    chip_adjust_learning_rate,
)

logger = logging.getLogger(__name__)


class Chip(BasePruning):

    def __init__(self, model, dataloaders, **CFG):
        self.dataloaders = dataloaders
        self.model = model
        self.CFG = CFG
        self.batch_size = self.CFG['batch_size']
        self.arch = self.CFG['arch']
        self.repeat = self.CFG['repeat']
        self.ci_dir = self.CFG['ci_dir']
        self.lr_type = self.CFG['lr_type']
        self.learning_rate = self.CFG['learning_rate']
        self.epochs = self.CFG['epochs']
        self.num_layers = self.CFG['num_layers']
        self.feature_map_dir = self.CFG['feature_map_dir']
        self.sparsity = self.CFG['sparsity']
        self.label_smooth = self.CFG['label_smooth']
        self.device = self.CFG['device']
        self.gpu = self.CFG['gpu']
        self.momentum = self.CFG['momentum']
        self.weight_decay = self.CFG['weight_decay']
        self.lr_decay_step = self.CFG['lr_decay_step']
        self.result_dir = self.CFG['result_dir']
        self.pretrain_dir = self.CFG['pretrain_dir']
        self.conv_index = torch.tensor([self.CFG['conv_index']])

        self.wandb_monitor = self.CFG.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.learning_rate}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/Chip', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/Chip/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.CFG}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet Chip', name=self.name)
            wandb.config.update(self.CFG)

    def get_feature_hook(self, module, input, output):
        os.makedirs('conv_feature_map/' + self.arch + '_repeat%d' %
                    (self.repeat),
                    exist_ok=True)
        np.save(
            'conv_feature_map/' + self.arch + '_repeat%d' % (self.repeat) +
            '/conv_feature_map_' + str(self.conv_index) + '.npy',
            output.cpu().numpy(),
        )
        self.conv_index += 1

    def inference(self):
        model = self.model
        model.eval()
        model.to(self.device)
        repeat = self.repeat
        with torch.no_grad():
            for batch_idx, (inputs,
                            targets) in enumerate(self.dataloaders['train']):
                # use 5 batches to get feature maps.
                if batch_idx >= repeat:
                    break

                inputs, targets = inputs.to(self.device), targets.to(
                    self.device)

                model(inputs)

    def reduced_1_row_norm(self, input, row_index, data_index):
        input[data_index, row_index, :] = torch.zeros(input.shape[-1])
        m = torch.norm(input[data_index, :, :], p='nuc').item()
        return m

    def ci_score(self, path_conv):
        conv_output = torch.tensor(np.round(np.load(path_conv), 4))
        conv_reshape = conv_output.reshape(conv_output.shape[0],
                                           conv_output.shape[1], -1)

        r1_norm = torch.zeros([conv_reshape.shape[0], conv_reshape.shape[1]])
        for i in range(conv_reshape.shape[0]):
            for j in range(conv_reshape.shape[1]):
                r1_norm[i, j] = self.reduced_1_row_norm(conv_reshape.clone(),
                                                        j,
                                                        data_index=i)

        ci = np.zeros_like(r1_norm)

        for i in range(r1_norm.shape[0]):
            original_norm = torch.norm(torch.tensor(conv_reshape[i, :, :]),
                                       p='nuc').item()
            ci[i] = original_norm - r1_norm[i]

        # return shape: [batch_size, filter_number]
        return ci

    def mean_repeat_ci(self, repeat, num_layers):
        layer_ci_mean_total = []
        for j in range(num_layers):
            print(f'Caclulating Mean Repeat CI for layer {j}')
            logger.info(f'Caclulating Mean Repeat CI for layer {j}')
            repeat_ci_mean = []
            for i in tqdm(range(repeat),
                          desc=f'Repeat CI for layer {j}',
                          total=repeat):
                index = j * repeat + i + 1
                # add
                path_conv = './conv_feature_map/{0}_repeat5/conv_feature_map_tensor({1}).npy'.format(
                    str(self.arch), str(index))

                batch_ci = self.ci_score(path_conv)
                single_repeat_ci_mean = np.mean(batch_ci, axis=0)
                repeat_ci_mean.append(single_repeat_ci_mean)

            layer_ci_mean = np.mean(repeat_ci_mean, axis=0)
            layer_ci_mean_total.append(layer_ci_mean)

        return np.array(layer_ci_mean_total)

    def load_resnet_model(self, model, oristate_dict):
        if len(self.gpu) > 1:
            name_base = 'module.'
        else:
            name_base = ''
        cfg = {
            'resnet_50': [3, 4, 6, 3],
        }

        state_dict = model.state_dict()

        current_cfg = cfg[self.arch]
        last_select_index = None

        all_honey_conv_weight = []

        bn_part_name = [
            '.weight',
            '.bias',
            '.running_mean',
            '.running_var',
        ]  # ,'.num_batches_tracked']
        prefix = self.ci_dir + '/ci_conv'
        subfix = '.npy'
        cnt = 1

        conv_weight_name = 'conv1.weight'
        all_honey_conv_weight.append(conv_weight_name)
        oriweight = oristate_dict[conv_weight_name]
        curweight = state_dict[name_base + conv_weight_name]
        orifilter_num = oriweight.size(0)
        currentfilter_num = curweight.size(0)

        if orifilter_num != currentfilter_num:
            logger.info('loading ci from: ' + prefix + str(cnt) + subfix)
            ci = np.load(prefix + str(cnt) + subfix)
            select_index = np.argsort(ci)[
                orifilter_num - currentfilter_num:]  # preserved filter id
            select_index.sort()

            for index_i, i in enumerate(select_index):
                state_dict[name_base + conv_weight_name][
                    index_i] = oristate_dict[conv_weight_name][i]
                for bn_part in bn_part_name:
                    state_dict[name_base + 'bn1' +
                               bn_part][index_i] = oristate_dict['bn1' +
                                                                 bn_part][i]

            last_select_index = select_index
        else:
            state_dict[name_base + conv_weight_name] = oriweight
            for bn_part in bn_part_name:
                state_dict[name_base + 'bn1' +
                           bn_part] = oristate_dict['bn1' + bn_part]

        state_dict[name_base + 'bn1' + '.num_batches_tracked'] = oristate_dict[
            'bn1' + '.num_batches_tracked']

        cnt += 1
        for layer, num in enumerate(current_cfg):
            layer_name = 'layer' + str(layer + 1) + '.'

            for k in range(num):
                iter = 3
                if k == 0:
                    iter += 1
                for l in range(iter):
                    record_last = True
                    if k == 0 and l == 2:
                        conv_name = layer_name + str(k) + '.downsample.0'
                        bn_name = layer_name + str(k) + '.downsample.1'
                        record_last = False
                    elif k == 0 and l == 3:
                        conv_name = layer_name + str(k) + '.conv' + str(l)
                        bn_name = layer_name + str(k) + '.bn' + str(l)
                    else:
                        conv_name = layer_name + str(k) + '.conv' + str(l + 1)
                        bn_name = layer_name + str(k) + '.bn' + str(l + 1)

                    conv_weight_name = conv_name + '.weight'
                    all_honey_conv_weight.append(conv_weight_name)
                    oriweight = oristate_dict[conv_weight_name]
                    curweight = state_dict[name_base + conv_weight_name]
                    orifilter_num = oriweight.size(0)
                    currentfilter_num = curweight.size(0)

                    if orifilter_num != currentfilter_num:
                        logger.info('loading ci from: ' + prefix + str(cnt) +
                                    subfix)
                        ci = np.load(prefix + str(cnt) + subfix)
                        select_index = np.argsort(
                            ci)[orifilter_num -
                                currentfilter_num:]  # preserved filter id
                        select_index.sort()

                        if last_select_index is not None:
                            for index_i, i in enumerate(select_index):
                                for index_j, j in enumerate(last_select_index):
                                    state_dict[name_base + conv_weight_name][
                                        index_i][index_j] = oristate_dict[
                                            conv_weight_name][i][j]

                                for bn_part in bn_part_name:
                                    state_dict[name_base + bn_name + bn_part][
                                        index_i] = oristate_dict[bn_name +
                                                                 bn_part][i]

                        else:
                            for index_i, i in enumerate(select_index):
                                state_dict[
                                    name_base +
                                    conv_weight_name][index_i] = oristate_dict[
                                        conv_weight_name][i]

                                for bn_part in bn_part_name:
                                    state_dict[name_base + bn_name + bn_part][
                                        index_i] = oristate_dict[bn_name +
                                                                 bn_part][i]

                        if record_last:
                            last_select_index = select_index

                    elif last_select_index is not None:
                        for index_i in range(orifilter_num):
                            for index_j, j in enumerate(last_select_index):
                                state_dict[name_base + conv_weight_name][
                                    index_i][index_j] = oristate_dict[
                                        conv_weight_name][index_i][j]

                        for bn_part in bn_part_name:
                            state_dict[name_base + bn_name +
                                       bn_part] = oristate_dict[bn_name +
                                                                bn_part]

                        if record_last:
                            last_select_index = None

                    else:
                        state_dict[name_base + conv_weight_name] = oriweight
                        for bn_part in bn_part_name:
                            state_dict[name_base + bn_name +
                                       bn_part] = oristate_dict[bn_name +
                                                                bn_part]
                        if record_last:
                            last_select_index = None

                    state_dict[name_base + bn_name +
                               '.num_batches_tracked'] = oristate_dict[
                                   bn_name + '.num_batches_tracked']
                    cnt += 1

        for name, module in model.named_modules():
            name = name.replace('module.', '')
            if isinstance(module, nn.Conv2d):
                conv_name = name + '.weight'
                if conv_name not in all_honey_conv_weight:
                    state_dict[name_base +
                               conv_name] = oristate_dict[conv_name]

            elif isinstance(module, nn.Linear):
                state_dict[name_base + name +
                           '.weight'] = oristate_dict[name + '.weight']
                state_dict[name_base + name + '.bias'] = oristate_dict[name +
                                                                       '.bias']

        model.load_state_dict(state_dict)

    def compress_model(self):
        model = self.model
        cov_layer = eval('model.maxpool')
        handler = cov_layer.register_forward_hook(self.get_feature_hook)
        self.inference()
        handler.remove()

        # ResNet50 per bottleneck
        for i in range(4):
            print(f'Calculating feature maps of ResNet50 for block {i}')
            logger.info(f'Calculating feature maps of ResNet50 for block {i}')
            block = eval('model.layer%d' % (i + 1))
            for j in tqdm(range(model.num_blocks[i]),
                          desc=f'Block {i}',
                          total=model.num_blocks[i]):
                cov_layer = block[j].relu1
                handler = cov_layer.register_forward_hook(
                    self.get_feature_hook)
                self.inference()
                handler.remove()

                cov_layer = block[j].relu2
                handler = cov_layer.register_forward_hook(
                    self.get_feature_hook)
                self.inference()
                handler.remove()

                cov_layer = block[j].relu3
                handler = cov_layer.register_forward_hook(
                    self.get_feature_hook)
                self.inference()
                handler.remove()

                if j == 0:
                    cov_layer = block[j].relu3
                    handler = cov_layer.register_forward_hook(
                        self.get_feature_hook)
                    self.inference()
                    handler.remove()

        repeat = self.repeat
        num_layers = self.num_layers
        save_path = os.path.join(self.save, 'CI_' + self.arch)
        ci = self.mean_repeat_ci(repeat, num_layers)

        if self.arch == 'resnet_50':
            num_layers = 53

        for i in range(num_layers):
            print(f'Saving CI of layer {i}')
            logger.info(f'Saving CI of layer {i}')
            os.makedirs(save_path, exist_ok=True)
            np.save(save_path + '/ci_conv{0}.npy'.format(str(i + 1)), ci[i])

        os.makedirs(self.result_dir, exist_ok=True)

        # save old training file
        now = datetime.datetime.now().strftime('%Y-%m-%d-%H:%M:%S')
        cp_file_dir = os.path.join(self.result_dir, 'cp_file/' + now)
        if os.path.exists(self.result_dir + '/model_best.pth.tar'):
            if not os.path.isdir(cp_file_dir):
                os.makedirs(cp_file_dir)
            shutil.copy(self.result_dir + '/config.txt', cp_file_dir)
            shutil.copy(self.result_dir + '/logger.log', cp_file_dir)
            shutil.copy(self.result_dir + '/model_best.pth.tar', cp_file_dir)
            shutil.copy(self.result_dir + '/checkpoint.pth.tar', cp_file_dir)

        cudnn.benchmark = True
        cudnn.enabled = True
        logger.info('args = %s', self)

        if self.sparsity:
            sparsity = extract_sparsity(self.sparsity)

        # load model
        logger.info('sparsity:' + str(sparsity))
        logger.info('==> Building model..')
        model = eval(self.arch)(sparsity=sparsity).cuda()
        logger.info(model)

        CLASSES = self.dataloaders['train'].dataset.classes
        print_freq = 128000 // self.batch_size
        criterion = nn.CrossEntropyLoss()
        criterion = criterion.cuda()
        criterion_smooth = CrossEntropyLabelSmooth(CLASSES, self.label_smooth)
        criterion_smooth = criterion_smooth.cuda()

        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=self.learning_rate,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )
        lr_decay_step = list(map(int, self.lr_decay_step.split(',')))
        # scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=lr_decay_step, gamma=0.1)

        start_epoch = 0
        best_top1_acc = 0
        best_top5_acc = 0

        # load the checkpoint if it exists
        checkpoint_dir = os.path.join(self.result_dir, 'checkpoint.pth.tar')

        logger.info('resuming from pretrain model')
        origin_model = eval(self.arch)(sparsity=[0.0] * 100).cuda()
        ckpt = torch.load(self.pretrain_dir)
        origin_model.load_state_dict(ckpt)
        oristate_dict = origin_model.state_dict()
        if self.arch == 'resnet_50':
            self.load_resnet_model(model, oristate_dict)
        else:
            raise

        # adjust the learning rate according to the checkpoint
        # for epoch in range(start_epoch):
        #     scheduler.step()

        # train the model
        scaler = GradScaler()
        epoch = start_epoch
        epochs_list = []
        train_top1_acc_list = []
        train_top5_acc_list = []
        val_top1_acc_list = []
        val_top5_acc_list = []
        while epoch < self.epochs:
            train_obj, train_top1_acc, train_top5_acc = self.train(
                epoch,
                self.dataloaders['train'],
                model,
                criterion_smooth,
                optimizer,
                scaler,
            )
            valid_obj, valid_top1_acc, valid_top5_acc = self.validate(
                epoch, self.dataloaders['val'], model, criterion)

            is_best = False
            if valid_top1_acc > best_top1_acc:
                best_top1_acc = valid_top1_acc
                best_top5_acc = valid_top5_acc
                is_best = True

            save_checkpoint(
                {
                    'epoch': epoch,
                    'state_dict': model.state_dict(),
                    'best_top1_acc': best_top1_acc,
                    'best_top5_acc': best_top5_acc,
                    'optimizer': optimizer.state_dict(),
                },
                is_best,
                self.result_dir,
            )

            epoch += 1

            if self.wandb_monitor:
                wandb.log({'best_top1_acc': best_top1_acc})

            epochs_list.append(epoch)
            train_top1_acc_list.append(train_top1_acc)
            train_top5_acc_list.append(train_top5_acc)
            val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
            val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

            df_data = np.array([
                epochs_list,
                train_top1_acc_list,
                train_top5_acc_list,
                val_top1_acc_list,
                val_top5_acc_list,
            ]).T
            df = pd.DataFrame(
                df_data,
                columns=[
                    'Epochs',
                    'Train Top1',
                    'Train Top5',
                    'Validation Top1',
                    'Validation Top5',
                ],
            )

            df.to_csv(
                f'{os.getcwd()}/logs/Chip/{self.name}.csv',
                index=False,
            )

    def train(self,
              epoch,
              train_loader,
              model,
              criterion,
              optimizer,
              scaler=None):
        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        model.train()
        end = time.time()

        epoch_iterator = tqdm(
            train_loader,
            desc=
            'Training network Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, targets) in enumerate(epoch_iterator):
            images = images.cuda()
            targets = targets.cuda()
            data_time.update(time.time() - end)

            chip_adjust_learning_rate(optimizer, epoch, i, len(train_loader))

            # compute output
            logits = model(images)
            loss = criterion(logits, targets)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(logits, targets, topk=(1, 5))
            n = images.size(0)
            losses.update(loss.item(), n)  # accumulated loss
            top1.update(prec1.item(), n)
            top5.update(prec5.item(), n)

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            logger.info(
                'Training network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                    'train_top1_acc': top1.val,
                    'train_top5_acc': top5.val,
                })

        return losses.avg, top1.avg, top5.avg

    def validate(self, epoch, val_loader, model, criterion):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            val_loader,
            desc=
            'Validating network Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model.eval()
        with torch.no_grad():
            end = time.time()
            for i, (images, targets) in enumerate(epoch_iterator):
                images = images.cuda()
                targets = targets.cuda()

                # compute output
                logits = model(images)
                loss = criterion(logits, targets)

                # measure accuracy and record loss
                pred1, pred5 = accuracy(logits, targets, topk=(1, 5))
                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

            logger.info(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import shutil
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules import loss
import torch.utils
from PIL import Image

__all__ = [
    'Lighting',
    'DistributionLoss',
    'adjust_learning_rate',
]

# lighting data augmentation
imagenet_pca = {
    'eigval':
    np.asarray([0.2175, 0.0188, 0.0045]),
    'eigvec':
    np.asarray([
        [-0.5675, 0.7192, 0.4009],
        [-0.5808, -0.0045, -0.8140],
        [-0.5836, -0.6948, 0.4203],
    ]),
}


class Lighting(object):

    def __init__(self,
                 alphastd,
                 eigval=imagenet_pca['eigval'],
                 eigvec=imagenet_pca['eigvec']):
        self.alphastd = alphastd
        assert eigval.shape == (3, )
        assert eigvec.shape == (3, 3)
        self.eigval = eigval
        self.eigvec = eigvec

    def __call__(self, img):
        if self.alphastd == 0.0:
            return img
        rnd = np.random.randn(3) * self.alphastd
        rnd = rnd.astype('float32')
        v = rnd
        old_dtype = np.asarray(img).dtype
        v = v * self.eigval
        v = v.reshape((3, 1))
        inc = np.dot(self.eigvec, v).reshape((3, ))
        img = np.add(img, inc)
        if old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        img = Image.fromarray(img.astype(old_dtype), 'RGB')
        return img

    def __repr__(self):
        return self.__class__.__name__ + '()'


class DistributionLoss(loss._Loss):
    """The KL-Divergence loss for the binary student model and real teacher
    output.

    output must be a pair of (model_output, real_output), both NxC tensors. The
    rows of real_output must all add up to one (probability scores); however,
    model_output must be the pre-softmax output of the network.
    """

    def forward(self, model_output, real_output):
        self.size_average = True

        if real_output.requires_grad:
            raise ValueError(
                'real network output should not require gradients.')

        model_output_log_prob = F.log_softmax(model_output, dim=1)
        real_output_soft = F.softmax(real_output, dim=1)
        del model_output, real_output

        real_output_soft = real_output_soft.unsqueeze(1)
        model_output_log_prob = model_output_log_prob.unsqueeze(2)

        cross_entropy_loss = -torch.bmm(real_output_soft,
                                        model_output_log_prob)
        if self.size_average:
            cross_entropy_loss = cross_entropy_loss.mean()
        else:
            cross_entropy_loss = cross_entropy_loss.sum()
        return cross_entropy_loss


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30
    epochs."""
    lr = args.lr * (0.1**(epoch // 30))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import torch
import torch.nn as nn
import scipy.optimize as optim
from itertools import count
from trailmet.utils import seed_everything
from trailmet.algorithms.quantize.quantize import (
    BaseQuantization,
    Conv2dFunctor,
    LinearFunctor,
)
from trailmet.algorithms.quantize.methods import (
    LearnedStepSizeQuantization,
    FixedClipValueQuantization,
)
from trailmet.algorithms.quantize.qmodel import (
    ParameterModuleWrapper,
    ActivationModuleWrapper,
)

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import AverageMeter, accuracy, save_checkpoint

logger = logging.getLogger(__name__)


class LAPQ(BaseQuantization):

    def __init__(self, model: nn.Module, dataloaders, **kwargs):
        super(LAPQ, self).__init__(**kwargs)
        self.model = model
        self.train_loader = dataloaders['train']
        self.test_loader = dataloaders['test']
        self.kwargs = kwargs
        self.w_bits = kwargs.get('W_BITS', 8)
        self.a_bits = kwargs.get('A_BITS', 8)
        self.calib_batches = kwargs.get('CALIB_BATCHES', 16)
        self.act_quant = kwargs.get('ACT_QUANT', True)
        self.test_before_calibration = kwargs.get('DRY_RUN', True)
        self.maxiter = kwargs.get('MAX_ITER', 1)
        self.maxfev = kwargs.get('MAX_FEV', 1)
        self.verbose = kwargs.get('VERBOSE', True)
        self.print_freq = kwargs.get('PRINT_FREQ', 20)
        self.gpu_id = kwargs.get('GPU_ID', 0)
        self.seed = kwargs.get('SEED', 42)
        seed_everything(self.seed)
        self.device = torch.device('cuda:{}'.format(self.gpu_id))
        if self.verbose:
            print('==> Using seed: {} and device: cuda:{}'.format(
                self.seed, self.gpu_id))
        self.calib_data = self.get_calib_samples(self.train_loader,
                                                 64 * self.calib_batches)
        self.eval_count = count(0)
        self.min_loss = 1e6

        self.wandb_monitor = self.kwargs.get('WANDB', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            f'{self.a_bits}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/LAPQ', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/LAPQ/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet LAPQ', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        self.model.to(self.device)
        self.search_absorbe_bn(self.model)
        args = {
            'bit_weights': self.w_bits,
            'bit_act': self.a_bits,
            'bcorr_w': True,
            'qtype': 'lp_norm',
            'lp': 2.0,
        }
        layers = []
        layers += [
            n for n, m in self.model.named_modules()
            if isinstance(m, nn.Conv2d)
        ][1:-1]
        if self.act_quant:
            layers += [
                n for n, m in self.model.named_modules()
                if isinstance(m, nn.ReLU)
            ][1:-1]
            layers += [
                n for n, m in self.model.named_modules()
                if isinstance(m, nn.ReLU6)
            ][1:-1]

        if self.test_before_calibration:
            args['qtype'] = 'max_static'
            cnn = copy.deepcopy(self.model)
            qm = QuantModel(cnn, args, layers)

            valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                qm.model, self.test_loader)

            print(
                '==> Quantization (W{}A{}) accuracy before LAPQ: {:.4f} | {:.4f}'
                .format(self.w_bits, self.a_bits, valid_top1_acc,
                        valid_top5_acc))
            logger.info(
                '==> Quantization (W{}A{}) accuracy before LAPQ: {:.4f} | {:.4f}'
                .format(self.w_bits, self.a_bits, valid_top1_acc,
                        valid_top5_acc))
            del qm, cnn

        ps = np.linspace(2, 4, 10)
        losses = []

        tk1 = tqdm(ps, total=len(ps))

        for p in tk1:
            args['qtype'] = 'lp_norm'
            args['lp'] = p
            cnn = copy.deepcopy(self.model)
            qm = QuantModel(cnn, args, layers)
            loss = self.evaluate_loss(model=qm.model, device=self.device)
            losses.append(loss.item())
            tk1.set_postfix(p_val=p, loss=loss.item())
            del qm, cnn
        # using quadratic interpolation to approximate the optimal quantization step size ∆p∗
        z = np.polyfit(ps, losses, 2)
        y = np.poly1d(z)
        p_intr = y.deriv().roots[0]

        print('==> using p intr : {:.2f}'.format(p_intr))
        logger.info('==> using p intr : {:.2f}'.format(p_intr))
        args['lp'] = p_intr
        quant_model = QuantModel(self.model, args, layers)

        valid_loss, valid_top1_acc, valid_top5_acc = self.test(
            quant_model.model, self.test_loader)
        lp_point = quant_model.get_clipping()

        print(
            '==> Quantization (W{}A{}) accuracy before Optimization: {:.4f} | {:.4f}'
            .format(self.w_bits, self.a_bits, valid_top1_acc, valid_top5_acc))
        print('==> Loss after LpNormQuantization: {:.4f}'.format(valid_loss))
        print('==> Starting Powell Optimization')

        logger.info(
            '==> Quantization (W{}A{}) accuracy before Optimization: {:.4f} | {:.4f}'
            .format(self.w_bits, self.a_bits, valid_top1_acc, valid_top5_acc))
        logger.info(
            '==> Loss after LpNormQuantization: {:.4f}'.format(valid_loss))
        logger.info('==> Starting Powell Optimization')

        min_method = 'Powell'
        min_options = {'maxiter': self.maxiter, 'maxfev': self.maxfev}
        init_scale = lp_point.cpu().numpy()
        count_iter = count(0)

        def local_search_callback(x):
            it = next(count_iter)
            quant_model.set_clipping(x, self.device)
            loss = self.evaluate_loss(quant_model.model, self.device)
            if self.verbose:
                print('\n==> Loss at end of iter [{}] : {:.4f}\n'.format(
                    it, loss.item()))

        self.pbar = tqdm(total=min(self.maxiter, self.maxfev))
        res = optim.minimize(
            lambda scales: self.evaluate_calibration(scales, quant_model, self.
                                                     device),
            init_scale,
            method=min_method,
            options=min_options,
            callback=local_search_callback,
        )
        self.pbar.close()
        scales = res.x
        print('==> Layer-wise Scales :\n', scales)
        logger.info(f'==> Layer-wise Scales :{scales}')

        quant_model.set_clipping(scales, self.device)

        valid_loss, valid_top1_acc, valid_top5_acc = self.test(
            quant_model.model, self.test_loader)
        print('==> Full quantization (W{}A{}) accuracy: {}'.format(
            self.w_bits, self.a_bits, valid_top1_acc))

        logger.info('==> Full quantization (W{}A{}) accuracy: {}'.format(
            self.w_bits, self.a_bits, valid_top1_acc))
        self.qnn = copy.deepcopy(quant_model.model)
        return self.qnn

    def test(self, model, dataloader, loss_fn=nn.CrossEntropyLoss()):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating network (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model.eval()
        model.to(self.device)

        with torch.no_grad():
            end = time.time()

            for i, (images, labels) in enumerate(epoch_iterator):
                images = images.to(self.device, dtype=torch.float)
                labels = labels.to(self.device)

                preds = model(images)

                loss = loss_fn(preds, labels)

                pred1, pred5 = accuracy(preds, labels, topk=(1, 5))

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))
        return losses.avg, top1.avg, top5.avg

    def evaluate_calibration(self, scales, QM, device):
        eval_count = next(self.eval_count)
        QM.set_clipping(scales, device)
        loss = self.evaluate_loss(QM.model, device).item()
        if loss < self.min_loss:
            self.min_loss = loss
        # if self.verbose and eval_count%self.print_freq==0:
        #     print("==> iteration: {}, minimum loss so far: {:.4f}".format(
        #     eval_count, self.min_loss))
        self.pbar.set_postfix(curr_loss=loss, min_loss=self.min_loss)
        self.pbar.update(1)
        return loss

    def evaluate_loss(self, model: nn.Module, device):
        criterion = torch.nn.CrossEntropyLoss().to(device)
        model.eval()
        with torch.no_grad():
            if not hasattr(self, 'cal_set'):
                self.cal_set = []
                for i, (images, target) in enumerate(self.train_loader):
                    if (i >= self.calib_batches
                        ):  # TODO: make this robust for variable batch size
                        break
                    images = images.to(device, non_blocking=True)
                    target = target.to(device, non_blocking=True)
                    self.cal_set.append((images, target))
            res = torch.tensor([0.0]).to(device)
            for i in range(len(self.cal_set)):
                images, target = self.cal_set[i]
                output = model(images)
                loss = criterion(output, target)
                res += loss
            return res / len(self.cal_set)


class QuantModel:

    def __init__(self, model, args, quantizable_layers, optimizer_bridge=None):
        self.model = model
        self.args = args
        self.bit_weights = args['bit_weights']
        self.bit_act = args['bit_act']
        self.post_relu = True

        self.replacement_factory = {
            nn.ReLU: ActivationModuleWrapper,
            nn.ReLU6: ActivationModuleWrapper,
            nn.Conv2d: ParameterModuleWrapper,
        }
        self.functor_map = {
            nn.Conv2d: Conv2dFunctor,
            nn.Linear: LinearFunctor,
        }
        self.optimizer_bridge = optimizer_bridge

        self.quantization_wrappers = []
        self.quantizable_modules = []
        self.quantizable_layers = quantizable_layers
        self._pre_process_container(model)
        self._create_quantization_wrappers()
        self.quantization_params = LearnedStepSizeQuantization.learned_parameters(
        )

    def load_state_dict(self, state_dict):
        for name, qwrapper in self.quantization_wrappers:
            qwrapper.load_state_dict(state_dict)

    def freeze(self):
        for n, p in self.model.named_parameters():
            # TODO: hack, make it more robust
            if not np.any([qp in n for qp in self.quantization_params]):
                p.requires_grad = False

    @staticmethod
    def has_children(module):
        try:
            next(module.children())
            return True
        except StopIteration:
            return False

    def _create_quantization_wrappers(self):
        for qm in self.quantizable_modules:
            # replace module by it's wrapper
            fn = (self.functor_map[type(qm.module)](qm.module)
                  if type(qm.module) in self.functor_map else None)
            args = {
                'bits_out': self.bit_act,
                'bits_weight': self.bit_weights,
                'forward_functor': fn,
                'post_relu': self.post_relu,
                'optim_bridge': self.optimizer_bridge,
            }
            args.update(self.args)
            if hasattr(qm, 'bn'):
                args['bn'] = qm.bn
            module_wrapper = self.replacement_factory[type(qm.module)](
                qm.full_name, qm.module, **args)
            setattr(qm.container, qm.name, module_wrapper)
            self.quantization_wrappers.append((qm.full_name, module_wrapper))

    def _pre_process_container(self, container, prefix=''):
        prev, prev_name = None, None
        for name, module in container.named_children():
            # if is_bn(module) and is_absorbing(prev) and prev_name in self.quantizable_layers:
            #     # Pass BN module to prev module quantization wrapper for BN folding/unfolding
            #     self.quantizable_modules[-1].bn = module

            full_name = prefix + name
            if full_name in self.quantizable_layers:
                self.quantizable_modules.append(
                    type(
                        '',
                        (object, ),
                        {
                            'name': name,
                            'full_name': full_name,
                            'module': module,
                            'container': container,
                        },
                    )())

            if self.has_children(module):
                # For container we call recursively
                self._pre_process_container(module, full_name + '.')

            prev = module
            prev_name = full_name

    def get_qwrappers(self):
        return [
            qwrapper for (name, qwrapper) in self.quantization_wrappers
            if qwrapper.__enabled__()
        ]

    def set_clipping(self, clipping,
                     device):  # TODO: handle device internally somehow
        qwrappers = self.get_qwrappers()
        for i, qwrapper in enumerate(qwrappers):
            qwrapper.set_quantization(
                FixedClipValueQuantization,
                {
                    'clip_value': clipping[i],
                    'device': device
                },
            )

    def get_clipping(self):
        clipping = []
        qwrappers = self.get_qwrappers()
        for i, qwrapper in enumerate(qwrappers):
            q = qwrapper.get_quantization()
            clip_value = getattr(q, 'alpha')
            clipping.append(clip_value.item())

        return qwrappers[0].get_quantization().alpha.new_tensor(clipping)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import math
import torch
import torch.nn as nn

# from .layers import ModuleInjection, PrunableBatchNorm2d
from .base_model import BaseModel
import numpy as np
"""MobileNetV2 in PyTorch.

See the paper "Inverted Residuals and Linear Bottlenecks: Mobile Networks for
Classification, Detection and Segmentation" for more details. Code is taken
from
https://github.com/kuangliu/pytorch-cifar/blob/master/models/mobilenetv2.py
"""
import torch
import torch.nn as nn
import torch.nn.functional as F


class HardBinaryConv(nn.Module):

    def __init__(self, in_chn, out_chn, kernel_size=3, stride=1, padding=1):
        super(HardBinaryConv, self).__init__()
        self.stride = stride
        self.padding = padding
        self.number_of_weights = in_chn * out_chn * kernel_size * kernel_size
        self.shape = (out_chn, in_chn, kernel_size, kernel_size)
        self.weights = nn.Parameter(torch.rand(
            (self.number_of_weights, 1)) * 0.001,
                                    requires_grad=True)

    def forward(self, x):
        real_weights = self.weights.view(self.shape)
        scaling_factor = torch.mean(
            torch.mean(
                torch.mean(torch.abs(real_weights), dim=3, keepdim=True),
                dim=2,
                keepdim=True,
            ),
            dim=1,
            keepdim=True,
        )
        # print(scaling_factor, flush=True)
        scaling_factor = scaling_factor.detach()
        binary_weights_no_grad = scaling_factor * torch.sign(real_weights)
        cliped_weights = torch.clamp(real_weights, -1.0, 1.0)
        binary_weights = (binary_weights_no_grad.detach() -
                          cliped_weights.detach() + cliped_weights)
        #         print(binary_weights.shape)
        y = F.conv2d(x,
                     binary_weights,
                     stride=self.stride,
                     padding=self.padding)

        return y


class BinaryActivation(nn.Module):

    def __init__(self):
        super(BinaryActivation, self).__init__()

    def forward(self, x):
        out_forward = torch.sign(x)
        # out_e1 = (x^2 + 2*x)
        # out_e2 = (-x^2 + 2*x)
        out_e_total = 0
        mask1 = x < -1
        mask2 = x < 0
        mask3 = x < 1
        out1 = (-1) * mask1.type(
            torch.float32) + (x * x + 2 * x) * (1 - mask1.type(torch.float32))
        out2 = out1 * mask2.type(
            torch.float32) + (-x * x + 2 * x) * (1 - mask2.type(torch.float32))
        out3 = out2 * mask3.type(
            torch.float32) + 1 * (1 - mask3.type(torch.float32))
        out = out_forward.detach() - out3.detach() + out3

        return out


class Block(nn.Module):
    """Expand + depthwise + pointwise."""

    def __init__(self, in_planes, out_planes, expansion, stride, binary=True):
        super(Block, self).__init__()
        self.stride = stride
        self.binary = binary
        planes = expansion * in_planes

        if not self.binary:
            self.act = F.relu
            self.conv1 = nn.Conv2d(in_planes,
                                   planes,
                                   kernel_size=1,
                                   stride=1,
                                   padding=0,
                                   bias=False)
            self.bn1 = nn.BatchNorm2d(planes)
            self.conv2 = nn.Conv2d(
                planes,
                planes,
                kernel_size=3,
                stride=stride,
                padding=1,
                groups=planes,
                bias=False,
            )
            self.bn2 = nn.BatchNorm2d(planes)
            self.conv3 = nn.Conv2d(planes,
                                   out_planes,
                                   kernel_size=1,
                                   stride=1,
                                   padding=0,
                                   bias=False)
            self.bn3 = nn.BatchNorm2d(out_planes)

            self.shortcut = nn.Sequential()
            if stride == 1 and in_planes != out_planes:
                conv_module = nn.Conv2d(
                    in_planes,
                    out_planes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                )
                bn_module = nn.BatchNorm2d(out_planes)
                self.shortcut = nn.Sequential(conv_module, bn_module)
        else:
            self.act = BinaryActivation()
            self.conv1 = HardBinaryConv(in_planes,
                                        planes,
                                        kernel_size=1,
                                        stride=1,
                                        padding=0)
            self.bn1 = nn.BatchNorm2d(planes)
            self.conv2 = HardBinaryConv(planes,
                                        planes,
                                        kernel_size=3,
                                        stride=stride,
                                        padding=1)
            self.bn2 = nn.BatchNorm2d(planes)
            self.conv3 = HardBinaryConv(planes,
                                        out_planes,
                                        kernel_size=1,
                                        stride=1,
                                        padding=0)
            self.bn3 = nn.BatchNorm2d(out_planes)

            self.shortcut = nn.Sequential()
            if stride == 1 and in_planes != out_planes:
                conv_module = nn.Conv2d(
                    in_planes,
                    out_planes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                )
                bn_module = nn.BatchNorm2d(out_planes)
                self.shortcut = nn.Sequential(conv_module, bn_module)

    def forward(self, x):
        out = self.act(self.bn1(self.conv1(x)))
        out = self.act(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out = out + self.shortcut(x) if self.stride == 1 else out
        return out


class MobileNetv2(BaseModel):
    # (expansion, out_planes, num_blocks, stride)
    cfg = [
        (1, 16, 1, 1),
        (6, 24, 2, 2),  # NOTE: change stride 2 -> 1 for CIFAR10
        (6, 32, 3, 2),
        (6, 64, 4, 2),
        (6, 96, 3, 1),
        (6, 160, 3, 1),
        (6, 320, 1, 1),
    ]

    def __init__(self, num_classes=10, num_fp=0):
        super(MobileNetv2, self).__init__()
        # NOTE: change conv1 stride 2 -> 1 for CIFAR10
        self.conv1 = nn.Conv2d(3,
                               32,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        self.num_fp = num_fp
        self.curr_fp = 0
        self.layers = self._make_layers(in_planes=32)
        self.conv2 = nn.Conv2d(320,
                               1280,
                               kernel_size=1,
                               stride=2,
                               padding=0,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(1280)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)
        self.linear = nn.Linear(1280, num_classes)

    def _make_layers(self, in_planes):
        layers = []
        for expansion, out_planes, num_blocks, stride in self.cfg:
            strides = [stride] + [1] * (num_blocks - 1)
            if self.curr_fp < self.num_fp:
                for stride in strides:
                    layers.append(
                        Block(in_planes,
                              out_planes,
                              expansion,
                              stride,
                              binary=False))
                    in_planes = out_planes
                    self.curr_fp += 1
            else:
                for stride in strides:
                    layers.append(
                        Block(in_planes,
                              out_planes,
                              expansion,
                              stride,
                              binary=True))
                    in_planes = out_planes
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layers(out)
        out = F.relu(self.bn2(self.conv2(out)))
        # NOTE: change pooling kernel_size 7 -> 4 for CIFAR10
        #         out = F.avg_pool2d(out, 7)
        out = self.avgpool(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def get_mobilenet(num_classes, num_fp=0):
    """Returns the requested model, ready for training/pruning with the
    specified method.

    :param model: str
    :param method: full or prune
    :param num_classes: int, num classes in the dataset
    :return: A prunable MobileNet model
    """
    net = MobileNetv2(num_classes, num_fp)
    return net



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from trailmet.algorithms.distill.distill import Distillation
from trailmet.algorithms.distill.losses import KDTransferLoss

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import AverageMeter, accuracy, save_checkpoint, seed_everything

logger = logging.getLogger(__name__)

# Hinton's Knowledge Distillation
seed_everything(43)


class KDTransfer(Distillation):
    """Class to compress model using distillation via kd transfer."""

    def __init__(self, teacher_model, student_model, dataloaders, **kwargs):
        super(KDTransfer, self).__init__(**kwargs)
        self.teacher_model = teacher_model
        self.student_model = student_model
        self.dataloaders = dataloaders
        self.kwargs = kwargs
        self.lambda_ = self.kwargs['DISTILL_ARGS'].get('LAMBDA', 0.5)
        self.temperature = self.kwargs['DISTILL_ARGS'].get('TEMPERATURE', 5)
        self.ce_loss = nn.CrossEntropyLoss()
        self.kd_loss = KDTransferLoss(self.temperature, 'batchmean')

        self.epochs = kwargs['DISTILL_ARGS'].get('EPOCHS', 200)
        self.lr = kwargs['DISTILL_ARGS'].get('LR', 0.1)

        self.wandb_monitor = self.kwargs.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.lr}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/Response_KD', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/Response_KD/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet Response_KD', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        """Function to transfer knowledge from teacher to student."""
        # include teacher training options
        self.distill(
            self.teacher_model,
            self.student_model,
            self.dataloaders,
            **self.kwargs['DISTILL_ARGS'],
        )

    def distill(self, teacher_model, student_model, dataloaders, **kwargs):
        print('=====> TRAINING STUDENT NETWORK <=====')
        logger.info('=====> TRAINING STUDENT NETWORK <=====')
        test_only = kwargs.get('TEST_ONLY', False)
        lr = kwargs.get('LR', 0.1)
        weight_decay = kwargs.get('WEIGHT_DECAY', 0.0005)

        optimizer = torch.optim.SGD(
            student_model.parameters(),
            lr=lr,
            weight_decay=weight_decay,
            momentum=0.9,
            nesterov=True,
        )

        milestones = kwargs.get('MILESTONES', [60, 120, 160])
        gamma = kwargs.get('GAMMA', 0.2)

        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer,
                                                         milestones=milestones,
                                                         gamma=gamma,
                                                         verbose=False)

        criterion = self.criterion
        best_top1_acc = 0

        if test_only == False:
            epochs_list = []
            val_top1_acc_list = []
            val_top5_acc_list = []
            for epoch in range(self.epochs):
                t_loss = self.train_one_epoch(
                    teacher_model,
                    student_model,
                    dataloaders['train'],
                    criterion,
                    optimizer,
                    epoch,
                )

                valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                    teacher_model, student_model, dataloaders['val'],
                    criterion, epoch)

                # use conditions for different schedulers e.g. ReduceLROnPlateau needs scheduler.step(v_loss)
                scheduler.step()

                is_best = False
                if valid_top1_acc > best_top1_acc:
                    best_top1_acc = valid_top1_acc
                    is_best = True

                save_checkpoint(
                    {
                        'epoch': epoch,
                        'state_dict': student_model.state_dict(),
                        'best_top1_acc': best_top1_acc,
                        'optimizer': optimizer.state_dict(),
                        'scheduler': scheduler.state_dict(),
                    },
                    is_best,
                    self.save,
                )

                if self.wandb_monitor:
                    wandb.log({'best_top1_acc': best_top1_acc})

                epochs_list.append(epoch)
                val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
                val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

                df_data = np.array([
                    epochs_list,
                    val_top1_acc_list,
                    val_top5_acc_list,
                ]).T
                df = pd.DataFrame(
                    df_data,
                    columns=[
                        'Epochs',
                        'Validation Top1',
                        'Validation Top5',
                    ],
                )
                df.to_csv(f'{os.getcwd()}/logs/Response_KD/{self.name}.csv',
                          index=False)

    def train_one_epoch(self, teacher_model, student_model, dataloader,
                        loss_fn, optimizer, epoch):
        teacher_model.eval()
        student_model.train()

        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')

        end = time.time()

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Training student network Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, labels) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.to(self.device, dtype=torch.float)
            labels = labels.to(self.device)

            teacher_preds = teacher_model(images)
            student_preds = student_model(images)
            loss = loss_fn(teacher_preds, student_preds, labels)
            n = images.size(0)
            losses.update(loss.item(), n)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            logger.info(
                'Training student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                })

        return losses.avg

    def test(self, teacher_model, student_model, dataloader, loss_fn, epoch):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating student network Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        teacher_model.eval()
        student_model.eval()

        with torch.no_grad():
            end = time.time()

            for i, (images, labels) in enumerate(epoch_iterator):
                images = images.to(self.device, dtype=torch.float)
                labels = labels.to(self.device)

                teacher_preds = teacher_model(images)
                student_preds = student_model(images)
                loss = loss_fn(teacher_preds, student_preds, labels)

                pred1, pred5 = accuracy(student_preds, labels, topk=(1, 5))

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))
        return losses.avg, top1.avg, top5.avg

    def criterion(self, out_t, out_s, labels):
        ce_loss = self.ce_loss(out_s, labels)
        kd_loss = self.kd_loss(out_t, out_s)
        return (self.lambda_ * ce_loss + (1 - self.lambda_) *
                (self.temperature**2) * kd_loss)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from trailmet.algorithms.binarize.binarize import BaseBinarize
from trailmet.models.resnet import ResNetCifar
from trailmet.models.resnet import ResNet
import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import DistributionLoss
from trailmet.utils import AverageMeter, save_checkpoint, accuracy

import os
import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import time
import numpy as np

logger = logging.getLogger(__name__)


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=stride,
                     padding=1,
                     bias=False)


class BinaryActivation(nn.Module):

    def __init__(self):
        super(BinaryActivation, self).__init__()

    def forward(self, x):
        out_forward = torch.sign(x)
        # out_e1 = (x^2 + 2*x)
        # out_e2 = (-x^2 + 2*x)
        out_e_total = 0
        mask1 = x < -1
        mask2 = x < 0
        mask3 = x < 1
        out1 = (-1) * mask1.type(
            torch.float32) + (x * x + 2 * x) * (1 - mask1.type(torch.float32))
        out2 = out1 * mask2.type(
            torch.float32) + (-x * x + 2 * x) * (1 - mask2.type(torch.float32))
        out3 = out2 * mask3.type(
            torch.float32) + 1 * (1 - mask3.type(torch.float32))
        out = out_forward.detach() - out3.detach() + out3

        return out


class LearnableBias(nn.Module):

    def __init__(self, out_chn):
        super(LearnableBias, self).__init__()
        self.bias = nn.Parameter(torch.zeros(1, out_chn, 1, 1),
                                 requires_grad=True)

    def forward(self, x):
        out = x + self.bias.expand_as(x)
        return out


class HardBinaryConv(nn.Module):

    def __init__(self, in_chn, out_chn, kernel_size=3, stride=1, padding=1):
        super(HardBinaryConv, self).__init__()
        self.stride = stride
        self.padding = padding
        self.number_of_weights = in_chn * out_chn * kernel_size * kernel_size
        self.shape = (out_chn, in_chn, kernel_size, kernel_size)
        # self.weight = nn.Parameter(torch.rand((self.number_of_weights,1)) * 0.001, requires_grad=True)
        self.weight = nn.Parameter(torch.rand((self.shape)) * 0.001,
                                   requires_grad=True)

    def forward(self, x):
        # real_weights = self.weights.view(self.shape)
        real_weights = self.weight
        scaling_factor = torch.mean(
            torch.mean(torch.mean(abs(real_weights), dim=3, keepdim=True),
                       dim=2,
                       keepdim=True),
            dim=1,
            keepdim=True,
        )
        # print(scaling_factor, flush=True)
        scaling_factor = scaling_factor.detach()
        binary_weights_no_grad = scaling_factor * torch.sign(real_weights)
        cliped_weights = torch.clamp(real_weights, -1.0, 1.0)
        binary_weights = (binary_weights_no_grad.detach() -
                          cliped_weights.detach() + cliped_weights)
        # print(binary_weights, flush=True)
        y = F.conv2d(x,
                     binary_weights,
                     stride=self.stride,
                     padding=self.padding)

        return y


class BasicBlock1(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock1, self).__init__()

        self.move0 = LearnableBias(inplanes)
        self.binary_activation = BinaryActivation()
        self.binary_conv = conv3x3(inplanes, planes, stride=stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.move1 = LearnableBias(planes)
        self.prelu = nn.PReLU(planes)
        self.move2 = LearnableBias(planes)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.move0(x)
        out = self.binary_activation(out)
        out = self.binary_conv(out)
        out = self.bn1(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.move1(out)
        out = self.prelu(out)
        out = self.move2(out)

        return out


class BasicBlock2(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock2, self).__init__()

        self.move0 = LearnableBias(inplanes)
        self.binary_activation = BinaryActivation()
        self.binary_conv = HardBinaryConv(inplanes, planes, stride=stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.move1 = LearnableBias(planes)
        self.prelu = nn.PReLU(planes)
        self.move2 = LearnableBias(planes)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.move0(x)
        out = self.binary_activation(out)
        out = self.binary_conv(out)
        out = self.bn1(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.move1(out)
        out = self.prelu(out)
        out = self.move2(out)

        return out


class ReActNet(BaseBinarize):

    def __init__(self, teacher, model, dataloaders, **kwargs):
        super(ReActNet, self).__init__(**kwargs)
        self.teacher = teacher
        self.model = model
        self.layers = model.layers_size
        self.device = torch.device(
            'cuda:0' if torch.cuda.is_available() else 'cpu')
        self.dataloaders = dataloaders
        self.num_classes = model.num_classes
        self.insize = model.insize
        self.kwargs = kwargs

        self.batch_size1 = self.kwargs['ReActNet1_ARGS'].get('batch_size', 128)
        self.epochs1 = self.kwargs['ReActNet1_ARGS'].get('epochs', 128)
        self.learning_rate1 = self.kwargs['ReActNet1_ARGS'].get(
            'learning_rate', 2.5e-3)
        self.momentum1 = self.kwargs['ReActNet1_ARGS'].get('momentum', 0.9)
        self.weight_decay1 = self.kwargs['ReActNet1_ARGS'].get(
            'weight_decay', 1e-5)
        self.label_smooth1 = self.kwargs['ReActNet1_ARGS'].get(
            'label_smooth', 0.1)

        self.batch_size2 = self.kwargs['ReActNet2_ARGS'].get('batch_size', 128)
        self.epochs2 = self.kwargs['ReActNet2_ARGS'].get('epochs', 128)
        self.learning_rate2 = self.kwargs['ReActNet2_ARGS'].get(
            'learning_rate', 2.5e-3)
        self.momentum2 = self.kwargs['ReActNet2_ARGS'].get('momentum', 0.9)
        self.weight_decay2 = self.kwargs['ReActNet2_ARGS'].get(
            'weight_decay', 0)
        self.label_smooth2 = self.kwargs['ReActNet2_ARGS'].get(
            'label_smooth', 0.1)

        self.wandb_monitor = self.kwargs.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs1}',
            f'{self.learning_rate1}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/ReActNet', exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/ReActNet/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet ReActNet', name=self.name)
            wandb.config.update(self.kwargs)

    def make_model1(self):
        self.layers = [i * 2 for i in self.layers]

        if len(self.layers) == 3:
            new_model = ResNetCifar(
                BasicBlock1,
                self.layers,
                width=1,
                num_classes=self.num_classes,
                insize=self.insize,
            )
        else:
            new_model = ResNet(
                BasicBlock1,
                self.layers,
                width=1,
                num_classes=self.num_classes,
                insize=self.insize,
            )

        return new_model

    def make_model2(self):
        self.layers = [i * 2 for i in self.layers]

        if len(self.layers) == 3:
            new_model = ResNetCifar(
                BasicBlock2,
                self.layers,
                width=1,
                num_classes=self.num_classes,
                insize=self.insize,
            )
        else:
            new_model = ResNet(
                BasicBlock2,
                self.layers,
                width=1,
                num_classes=self.num_classes,
                insize=self.insize,
            )

        return new_model

    def train_one_epoch(self, model, teacher, scheduler, criterion, optimizer,
                        epoch):
        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        end = time.time()
        scheduler.step()

        train_loader = self.dataloaders['train']

        epoch_iterator = tqdm(
            train_loader,
            desc=
            'Training Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, target) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.cuda()
            target = target.cuda()

            # compute outputy
            logits_student = model(images)
            logits_teacher = teacher(images)
            loss = criterion(logits_student, logits_teacher)

            prec1, prec5 = accuracy(logits_student, target, topk=(1, 5))
            n = images.size(0)
            losses.update(loss.item(), n)  # accumulated loss
            top1.update(prec1.item(), n)
            top5.update(prec5.item(), n)

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            logger.info(
                'Training Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                    'train_top1_acc': top1.val,
                    'train_top5_acc': top5.val,
                })

        return losses.avg, top1.avg, top5.avg

    def base_train(self, model, teacher, epochs, criterion, scheduler,
                   optimizer, step):
        model.train()
        teacher = teacher.eval()
        for param in teacher.parameters():
            param.requires_grad = False
        epoch = 0
        epochs_list = []
        train_top1_acc_list = []
        train_top5_acc_list = []
        val_top1_acc_list = []
        val_top5_acc_list = []
        while epoch < epochs:
            train_obj, train_top1_acc, train_top5_acc = self.train_one_epoch(
                model, teacher, scheduler, criterion, optimizer, epoch)

            train_top1_acc_list.append(train_top1_acc)
            train_top5_acc_list.append(train_top5_acc)

            valid_loss, valid_top1_acc, valid_top5_acc = self.validate(
                epoch, self.dataloaders['val'], model, criterion)

            epochs_list.append(epoch)
            val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
            val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

            print.info(
                'Average Train Loss-{0}\tTop-1 Train Accuracy-{1}\tTop-5 Train Accuracy-{2}\tTop1 Validation Accuracy-{3}\tTop5 Validation Accuracy-{4}\tValidation Loss-{5}'
                .format(
                    train_obj,
                    train_top1_acc,
                    train_top5_acc,
                    valid_top1_acc,
                    valid_top5_acc,
                    valid_loss,
                ))

            logger.info(
                'Average Train Loss-{0}\tTop-1 Train Accuracy-{1}\tTop-5 Train Accuracy-{2}\tTop1 Validation Accuracy-{3}\tTop5 Validation Accuracy-{4}\tValidation Loss-{5}'
                .format(
                    train_obj,
                    train_top1_acc,
                    train_top5_acc,
                    valid_top1_acc,
                    valid_top5_acc,
                    valid_loss,
                ))
            epoch += 1

        df_data = np.array([
            epochs_list,
            train_top1_acc_list,
            train_top5_acc_list,
            val_top1_acc_list,
            val_top5_acc_list,
        ]).T
        df = pd.DataFrame(
            df_data,
            columns=[
                'Epochs',
                'Train Top1',
                'Train Top5',
                'Validation Top1',
                'Validation Top5',
            ],
        )
        df.to_csv(f'{os.getcwd()}/logs/ReActNet/Step-{step}-{self.name}.csv',
                  index=False)

        return model

    def train1(self):
        print('Step-1 Training with activations binarized for {} epochs\n'.
              format(self.epochs1))
        logger.info(
            'Step-1 Training with activations binarized for {} epochs\n'.
            format(self.epochs1))
        model = self.make_model1()
        model = model.to(self.device)
        teacher = self.teacher.to(self.device)

        all_parameters = model.parameters()
        weight_parameters = []
        for pname, p in model.named_parameters():
            if p.ndimension() == 4 or 'conv' in pname:
                weight_parameters.append(p)
        weight_parameters_id = list(map(id, weight_parameters))
        other_parameters = list(
            filter(lambda p: id(p) not in weight_parameters_id,
                   all_parameters))

        criterion = DistributionLoss()
        optimizer = torch.optim.Adam(
            [
                {
                    'params': other_parameters
                },
                {
                    'params': weight_parameters,
                    'weight_decay': self.weight_decay1
                },
            ],
            lr=self.learning_rate1,
        )
        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lambda step: (1.0 - step / self.epochs1), last_epoch=-1)

        model = self.base_train(model, teacher, self.epochs1, criterion,
                                scheduler, optimizer, '1')
        return model

    def train2(self, model):
        print(
            'Step-2 Training with both activations and weights binarized for {} epochs'
            .format(self.epochs2))
        logger.info(
            'Step-2 Training with both activations and weights binarized for {} epochs'
            .format(self.epochs2))
        # model = model.to(self.device)
        teacher = self.teacher.to(self.device)

        all_parameters = model.parameters()
        weight_parameters = []
        for pname, p in model.named_parameters():
            if p.ndimension() == 4 or 'conv' in pname:
                weight_parameters.append(p)
        weight_parameters_id = list(map(id, weight_parameters))
        other_parameters = list(
            filter(lambda p: id(p) not in weight_parameters_id,
                   all_parameters))

        criterion = DistributionLoss()
        optimizer = torch.optim.Adam(
            [
                {
                    'params': other_parameters
                },
                {
                    'params': weight_parameters,
                    'weight_decay': self.weight_decay2
                },
            ],
            lr=self.learning_rate2,
        )
        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lambda step: (1.0 - step / self.epochs2), last_epoch=-1)

        model = self.base_train(model, teacher, self.epochs2, criterion,
                                scheduler, optimizer, '2')
        return model

    def validate(self, epoch, val_loader, model, criterion):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            val_loader,
            desc=
            'Validating Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        # switch to evaluation mode
        model.eval()
        with torch.no_grad():
            end = time.time()

            for i, (images, target) in enumerate(epoch_iterator):
                images = images.cuda()
                target = target.cuda()

                # compute output
                logits = model(images)
                loss = criterion(logits, target)

                # measure accuracy and record loss
                pred1, pred5 = accuracy(logits, target, topk=(1, 5))
                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg

    def compress_model(self):
        model1 = self.train1()
        model2 = self.make_model2()
        model2 = model2.to(self.device)
        model2.load_state_dict(model1.state_dict(), strict=False)

        fin_output = self.train2(model2)

        return fin_output



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import random
import os
import numpy as np
import torch
import shutil
import torch.nn as nn
import torch.optim as optim
import re
import math

__all__ = [
    'AverageMeter',
    'save_checkpoint',
    'accuracy',
    'seed_everything',
    'pdist',
    'CrossEntropyLabelSmooth',
    'adjust_learning_rate',
    'strlist_to_list',
    'get_optimizer',
    'lp_loss',
    'extract_sparsity',
    'chip_adjust_learning_rate',
]


def chip_adjust_learning_rate(self, optimizer, epoch, step, len_iter):
    if self.lr_type == 'step':
        factor = epoch // 125
        # if epoch >= 80:
        #     factor = factor + 1
        lr = self.learning_rate * (0.1**factor)

    elif self.lr_type == 'step_5':
        factor = epoch // 10
        if epoch >= 80:
            factor = factor + 1
        lr = self.learning_rate * (0.5**factor)

    elif self.lr_type == 'cos':  # cos without warm-up
        if self.epochs > 5:
            lr = (0.5 * self.learning_rate *
                  (1 + math.cos(math.pi * (epoch - 5) / (self.epochs - 5))))
        else:
            lr = self.learning_rate

    elif self.lr_type == 'exp':
        step = 1
        decay = 0.96
        lr = self.learning_rate * (decay**(epoch // step))

    elif self.lr_type == 'fixed':
        lr = self.learning_rate
    else:
        raise NotImplementedError

    if epoch < 5:
        lr = lr * float(1 + step + epoch * len_iter) / (5.0 * len_iter)

    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

    if step == 0:
        print('learning_rate: ' + str(lr))


def extract_sparsity(sparsity):
    cprate_str = sparsity
    cprate_str_list = cprate_str.split('+')
    pat_cprate = re.compile(r'\d+\.\d*')
    pat_num = re.compile(r'\*\d+')
    cprate = []
    for x in cprate_str_list:
        num = 1
        find_num = re.findall(pat_num, x)
        if find_num:
            assert len(find_num) == 1
            num = int(find_num[0].replace('*', ''))
        find_cprate = re.findall(pat_cprate, x)
        assert len(find_cprate) == 1
        cprate += [float(find_cprate[0])] * num

    return cprate


def lp_loss(pred, tgt, p=2.0, reduction='none'):
    """Loss function measured in Lp Norm."""
    if reduction == 'none':
        return (pred - tgt).abs().pow(p).sum(1).mean()
    else:
        return (pred - tgt).abs().pow(p).mean()


def seed_everything(seed):
    'sets the random seed to ensure reproducibility'
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True


def accuracy(output, target, topk=(1, )):
    """Computes the accuracy over the k top predictions for the specified
    values of k."""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.reshape(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


class AverageMeter(object):
    """Computes and stores the average and current value."""

    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


def save_checkpoint(state, is_best, save, file_name=None):
    if not os.path.exists(save):
        os.makedirs(save)
    if file_name is not None:
        filename = os.path.join(save, f'{file_name}.pth.tar')
    else:
        filename = os.path.join(save, 'checkpoint.pth.tar')
    torch.save(state, filename)
    if is_best:
        best_filename = os.path.join(save, 'model_best.pth.tar')
        shutil.copyfile(filename, best_filename)


def pdist(e, squared=False, eps=1e-12):
    e_square = e.pow(2).sum(dim=1)
    prod = e @ e.t()
    res = (e_square.unsqueeze(1) + e_square.unsqueeze(0) -
           2 * prod).clamp(min=eps)

    if not squared:
        res = res.sqrt()

    res = res.clone()
    res[range(len(e)), range(len(e))] = 0
    return res


# label smooth
class CrossEntropyLabelSmooth(nn.Module):

    def __init__(self, num_classes, epsilon):
        super(CrossEntropyLabelSmooth, self).__init__()
        self.num_classes = num_classes
        self.epsilon = epsilon
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, inputs, targets):
        log_probs = self.logsoftmax(inputs)
        targets = torch.zeros_like(log_probs).scatter_(1, targets.unsqueeze(1),
                                                       1)
        targets = (1 -
                   self.epsilon) * targets + self.epsilon / self.num_classes
        loss = (-targets * log_probs).mean(0).sum()
        return loss


def adjust_learning_rate(optimizer, epoch, num_epochs, scheduler_type, lr):
    """Sets the learning rate to the initial LR decayed by 2 every 30
    epochs."""
    if scheduler_type == 1:
        new_lr = lr * (0.5**(epoch // 30))
        for param_group in optimizer.param_groups:
            param_group['lr'] = new_lr
    else:
        if epoch in [num_epochs * 0.5, num_epochs * 0.75]:
            for param_group in optimizer.param_groups:
                param_group['lr'] *= 0.1


def strlist_to_list(sstr, ttype):
    """
    example:
    # self.args.stage_pr = [0, 0.3, 0.3, 0.3, 0, ]
    # self.args.skip_layers = ['1.0', '2.0', '2.3', '3.0', '3.5', ]
    turn these into a list of <ttype> (float or str or int etc.)
    """
    if not sstr:
        return sstr
    out = []
    sstr = sstr.strip()
    if sstr.startswith('[') and sstr.endswith(']'):
        sstr = sstr[1:-1]
    for x in sstr.split(','):
        x = x.strip()
        if x:
            x = ttype(x)
            out.append(x)
    return out


def get_optimizer(self, optimizer_name: str, model, lr, weight_decay):
    """Returns the optimizer with the given name."""
    if optimizer_name == 'SGD':
        optimizer = optim.SGD(model.parameters(),
                              lr=lr,
                              weight_decay=weight_decay)
    if optimizer_name == 'Adam':
        optimizer = optim.Adam(model.parameters(),
                               lr=lr,
                               weight_decay=weight_decay)
    else:
        raise ValueError('Unknown optimizer: %s' % optimizer_name)
    return optimizer



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import copy
import time
import numpy as np
import logging
from .meta_pruner import MetaPruner


class Pruner(MetaPruner):

    def __init__(self, model, args, logger, runner):
        super(Pruner, self).__init__(model, args, logger, runner)
        self.logger = logger

    def prune(self):
        self._get_kept_wg_L1()
        self._prune_and_build_new_model()

        if self.args.reinit:
            if self.args.reinit == 'orth':
                self.logger.info('==> Reinit model: orthogonal initialization')
                for module in self.model.modules():
                    if isinstance(module, (nn.Conv2d, nn.Linear)):
                        nn.init.orthogonal_(module.weight.data)

        return self.model
